import os


class MsDataset(object):
    @classmethod
    def load_core(cls, data_dir, data_set):
        wav_file = os.path.join(data_dir, data_set, "wav.scp")
        text_file = os.path.join(data_dir, data_set, "text")
        with open(wav_file) as f:
            wav_lines = f.readlines()
        with open(text_file) as  f:
            text_lines = f.readlines()
        data_list = []
        for wav_line, text_line in zip(wav_lines, text_lines):
            item = {}
            item["Audio:FILE"] = wav_line.strip().split()[-1]
            item["Text:LABEL"] = " ".join(text_line.strip().split()[1:])
            data_list.append(item)
        return data_list

    @classmethod
    def load(cls, dataset_name, namespace="speech_asr", train_set="train", dev_set="validation"):
        if os.path.exists(dataset_name):
            data_dir = dataset_name
            ds_dict = {}
            ds_dict["train"] = cls.load_core(data_dir, train_set)
            ds_dict["validation"] = cls.load_core(data_dir, dev_set)
            ds_dict["raw_data_dir"] = data_dir
            return ds_dict
        else:
            from modelscope.msdatasets import MsDataset
            ds_dict = MsDataset.load(dataset_name=dataset_name, namespace=namespace)
            return ds_dict



import logging

from dotenv import load_dotenv
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger
from apipeline.frames import AudioRawFrame, TextFrame

from achatbot.processors.speech.audio_save_processor import AudioSaveProcessor
from achatbot.processors.aggregators.user_audio_response import UserAudioResponseAggregator
from achatbot.cmd.bots.base_daily import DailyRoomBot
from achatbot.common.types import DailyParams
from achatbot.transports.daily import DailyTransport
from achatbot.cmd.bots import register_ai_room_bots
from achatbot.types.frames import PathAudioRawFrame, LLMGenedTokensFrame, BotSpeakingFrame
from .helper import get_step_audio2_processor, get_step_audio2_llm, get_token2wav


load_dotenv(override=True)


@register_ai_room_bots.register
class DailyStepAudio2AQAABot(DailyRoomBot):
    """
    - use daily audio stream(bytes) --> Step2 voice processor --> text/audio_bytes
    """

    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

        self.vad_analyzer = None
        self.audio_llm = None
        self.token2wav = None

    def load(self):
        self.vad_analyzer = self.get_vad_analyzer()
        self.audio_llm = get_step_audio2_llm(self._bot_config.voice_llm)
        self.token2wav = get_token2wav(self._bot_config.voice_llm)

    async def arun(self):
        assert self.vad_analyzer is not None
        assert self.audio_llm is not None

        self.params = DailyParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            vad_enabled=True,
            vad_analyzer=self.vad_analyzer,
            vad_audio_passthrough=True,
        )

        # src/processors/voice/step_audio2_processor.py
        self._voice_processor = get_step_audio2_processor(
            self._bot_config.voice_llm,
            session=self.session,
            audio_llm=self.audio_llm,
            token2wav=self.token2wav,
        )
        if hasattr(self._voice_processor, "stream_info"):
            stream_info = self._voice_processor.stream_info
            self.params.audio_out_sample_rate = stream_info["sample_rate"]
            self.params.audio_out_channels = stream_info["channels"]
        logging.info(f"params: {self.params}")

        transport = DailyTransport(
            self.args.room_url,
            self.args.token,
            self.args.bot_name,
            self.params,
        )

        user_audio_save_processor = None
        bot_speak_audio_save_processor = None
        if self._save_audio:
            user_audio_save_processor = AudioSaveProcessor(
                prefix_name="user_audio_aggr", pass_raw_audio=True
            )
            bot_speak_audio_save_processor = AudioSaveProcessor(
                prefix_name="bot_speak", pass_raw_audio=True
            )

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    UserAudioResponseAggregator(),
                    user_audio_save_processor,
                    FrameLogger(include_frame_types=[AudioRawFrame]),
                    self._voice_processor,
                    FrameLogger(
                        include_frame_types=[TextFrame, AudioRawFrame, LLMGenedTokensFrame]
                    ),
                    bot_speak_audio_save_processor,
                    # FrameLogger(include_frame_types=[BotSpeakingFrame]),
                    transport.output_processor(),  # BotSpeakingFrame
                ]
            ),
            params=PipelineParams(
                allow_interruptions=False,
                enable_metrics=False,
                send_initial_empty_metrics=False,
            ),
        )

        transport.add_event_handlers(
            "on_first_participant_joined",
            [self.on_first_participant_joined, self.on_first_participant_say_hi],
        )
        transport.add_event_handler("on_participant_left", self.on_participant_left)
        transport.add_event_handler("on_call_state_updated", self.on_call_state_updated)

        await PipelineRunner().run(self.task)

    async def on_first_participant_say_hi(self, transport: DailyTransport, participant):
        await self._voice_processor.say(
            "你好，我是一名助手，欢迎语音聊天!",
            temperature=0.7,
            max_new_tokens=1024,
            top_k=20,
            top_p=0.95,
            repetition_penalty=1.1,
        )



# Copyright (c) 2020 Mobvoi Inc. (authors: Binbin Zhang, Di Wu)
#               2024 Alibaba Inc (Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified from ESPnet(https://github.com/espnet/espnet)
"""Positonal Encoding Module."""

import math
from typing import Tuple, Union

import torch
import torch.nn.functional as F
import numpy as np


class EspnetRelPositionalEncoding(torch.nn.Module):
    """Relative positional encoding module (new implementation).

    Details can be found in https://github.com/espnet/espnet/pull/2816.

    See : Appendix B in https://arxiv.org/abs/1901.02860

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.

    """

    def __init__(self, d_model: int, dropout_rate: float, max_len: int = 5000):
        """Construct an PositionalEncoding object."""
        super(EspnetRelPositionalEncoding, self).__init__()
        self.d_model = d_model
        self.xscale = math.sqrt(self.d_model)
        self.dropout = torch.nn.Dropout(p=dropout_rate)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_len))

    def extend_pe(self, x: torch.Tensor):
        """Reset the positional encodings."""
        if self.pe is not None:
            # self.pe contains both positive and negative parts
            # the length of self.pe is 2 * input_len - 1
            if self.pe.size(1) >= x.size(1) * 2 - 1:
                if self.pe.dtype != x.dtype or self.pe.device != x.device:
                    self.pe = self.pe.to(dtype=x.dtype, device=x.device)
                return
        # Suppose `i` means to the position of query vecotr and `j` means the
        # position of key vector. We use position relative positions when keys
        # are to the left (i>j) and negative relative positions otherwise (i<j).
        pe_positive = torch.zeros(x.size(1), self.d_model)
        pe_negative = torch.zeros(x.size(1), self.d_model)
        position = torch.arange(0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe_positive[:, 0::2] = torch.sin(position * div_term)
        pe_positive[:, 1::2] = torch.cos(position * div_term)
        pe_negative[:, 0::2] = torch.sin(-1 * position * div_term)
        pe_negative[:, 1::2] = torch.cos(-1 * position * div_term)

        # Reserve the order of positive indices and concat both positive and
        # negative indices. This is used to support the shifting trick
        # as in https://arxiv.org/abs/1901.02860
        pe_positive = torch.flip(pe_positive, [0]).unsqueeze(0)
        pe_negative = pe_negative[1:].unsqueeze(0)
        pe = torch.cat([pe_positive, pe_negative], dim=1)
        self.pe = pe.to(device=x.device, dtype=x.dtype)

    def forward(self, x: torch.Tensor, offset: Union[int, torch.Tensor] = 0) \
            -> Tuple[torch.Tensor, torch.Tensor]:
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        self.extend_pe(x)
        x = x * self.xscale
        pos_emb = self.position_encoding(size=x.size(1), offset=offset)
        return self.dropout(x), self.dropout(pos_emb)

    def position_encoding(self,
                          offset: Union[int, torch.Tensor],
                          size: int) -> torch.Tensor:
        """ For getting encoding in a streaming fashion

        Attention!!!!!
        we apply dropout only once at the whole utterance level in a none
        streaming way, but will call this function several times with
        increasing input size in a streaming scenario, so the dropout will
        be applied several times.

        Args:
            offset (int or torch.tensor): start offset
            size (int): required size of position encoding

        Returns:
            torch.Tensor: Corresponding encoding
        """
        pos_emb = self.pe[
            :,
            self.pe.size(1) // 2 - size + 1: self.pe.size(1) // 2 + size,
        ]
        return pos_emb



import os
import modal

app = modal.App("vllm-generate")

vllm_image = (
    modal.Image.debian_slim(python_version="3.11")
    .pip_install(
        "vllm==0.7.3",
        "huggingface_hub[hf_transfer]==0.26.2",
        "flashinfer-python==0.2.0.post2",
        extra_index_url="https://flashinfer.ai/whl/cu124/torch2.5",
    )
    .env(
        {
            "HF_HUB_ENABLE_HF_TRANSFER": "1",
            "TORCH_CUDA_ARCH_LIST": "8.0 8.6 8.7 8.9 9.0",
        }
    )  # faster model transfers
)

hf_cache_vol = modal.Volume.from_name("huggingface-cache", create_if_missing=True)

# PP need close v1
vllm_image = vllm_image.env(
    {
        "VLLM_USE_V1": os.getenv("VLLM_USE_V1", "1"),
    }
)

achatbot_vllm_image = vllm_image.pip_install(
    "achatbot==0.0.9.post1",
    extra_index_url="https://pypi.org/simple/",
)

HF_MODEL_DIR = "/root/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
VLLM_CACHE_DIR = "/root/.cache/vllm"
vllm_cache_vol = modal.Volume.from_name("vllm-cache", create_if_missing=True)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=vllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        VLLM_CACHE_DIR: vllm_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
def run_sync():
    from vllm import LLM, EngineArgs

    engine = LLM(
        model="/root/models/Qwen/Qwen2.5-0.5B",
        # dtype="half",
        # enforce_eager=True,
    )

    sampling_params = engine.get_default_sampling_params()
    print(sampling_params)
    sampling_params.n = 1
    sampling_params.seed = 42
    sampling_params.max_tokens = 2
    sampling_params.temperature = 0.9
    sampling_params.top_p = 1.0
    sampling_params.top_k = 50
    sampling_params.min_p = 0.0
    # Penalizers
    sampling_params.repetition_penalty = 1.1
    sampling_params.min_tokens = 0
    outputs = engine.generate(
        "Hello, my name is",
        sampling_params=sampling_params,
    )
    print(outputs)
    i = 1
    for part in outputs:
        print(part)
        if i == 3:
            print("stopping")
            break
        i += 1


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=vllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        VLLM_CACHE_DIR: vllm_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_async():
    import asyncio
    import uuid

    from vllm import AsyncEngineArgs, AsyncLLMEngine, SamplingParams
    # from vllm.engine.async_llm_engine import AsyncEngineArgs, AsyncLLMEngine

    engine = AsyncLLMEngine.from_engine_args(
        AsyncEngineArgs(
            model="/root/models/Qwen/Qwen2.5-0.5B",
            # dtype="half",
            # enforce_eager=True,
        )
    )

    sampling_params = SamplingParams(
        n=1,
        seed=42,
        max_tokens=2,
        temperature=0.9,
        top_p=1.0,
        top_k=50,
        min_p=0.0,
        # Penalizers
        repetition_penalty=1.1,
        min_tokens=0,
    )
    print(sampling_params)

    outputs_generator = engine.generate(
        "Hello, my name is", sampling_params=sampling_params, request_id=str(uuid.uuid4().hex)
    )
    i = 1
    async for part in outputs_generator:
        print(part)
        """
        RequestOutput(request_id=a80d0cce8c774cbaada1ae0bb993ddf5, prompt='Hello, my name is', prompt_token_ids=[9707, 11, 847, 829, 374], encoder_prompt=None, encoder_prompt_token_ids=None, prompt_logprobs=None, outputs=[CompletionOutput(index=0, text=' David', token_ids=[6798], cumulative_logprob=None, logprobs=None, finish_reason=None, stop_reason=None)], finished=False, metrics=None, lora_request=None, num_cached_tokens=None, multi_modal_placeholders={})
        """
        if part.outputs:
            print(part.outputs[0].token_ids[-1])
        if i == 3:
            print("stopping")
            break
        i += 1


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=achatbot_vllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        VLLM_CACHE_DIR: vllm_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_achatbot_generator():
    import uuid
    import os
    import asyncio
    import time

    from achatbot.core.llm.vllm.generator import VllmGenerator, VllmEngineArgs, AsyncEngineArgs
    from achatbot.common.types import SessionCtx
    from achatbot.common.session import Session
    from transformers import AutoTokenizer, GenerationConfig

    model = "/root/models/Qwen/Qwen2.5-0.5B"
    generator = VllmGenerator(
        **VllmEngineArgs(serv_args=AsyncEngineArgs(model=model).__dict__).__dict__,
    )
    tokenizer = AutoTokenizer.from_pretrained(model)

    # async def run():
    generation_config = {}
    if os.path.exists(os.path.join(model, "generation_config.json")):
        generation_config = GenerationConfig.from_pretrained(
            model, "generation_config.json"
        ).to_dict()
    # async def run():
    prompt_cases = [
        {"prompt": "hello, my name is", "kwargs": {"max_new_tokens": 30, "stop_ids": []}},
        {
            "prompt": "hello, my name is",
            "kwargs": {"max_new_tokens": 30, "stop_ids": [13]},
        },  # prefill cache token test
        {"prompt": "hello, what your name?", "kwargs": {"max_new_tokens": 30, "stop_ids": [13]}},
    ]

    # test the same session
    # session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    for case in prompt_cases:
        session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
        session.ctx.state["token_ids"] = tokenizer.encode("hello, my name is")
        tokens = tokenizer(case["prompt"])
        session.ctx.state["token_ids"] = tokens["input_ids"]
        # gen_kwargs = {**generation_config, **case["kwargs"], **tokens} # hack test, vllm have some bug :)
        gen_kwargs = {**case["kwargs"], **tokens}
        print("gen_kwargs:", gen_kwargs)
        first = True
        start_time = time.perf_counter()
        gen_texts = ""
        async for token_id in generator.generate(session, **gen_kwargs):
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate TTFT time: {ttft} s")
                first = False
            gen_text = tokenizer.decode(token_id)
            gen_texts += gen_text
            print(session.ctx.client_id, token_id, gen_text)
        print(session.ctx.client_id, gen_texts)

    # asyncio.run(run())


"""
modal run src/llm/vllm/examples/generate.py::run_sync (no stream)
modal run src/llm/vllm/examples/generate.py::run_async (stream)
modal run src/llm/vllm/examples/generate.py::run_achatbot_generator (stream)
"""


# @app.local_entrypoint()
# def main():
#    run_sync.remote()



from abc import ABC, abstractmethod
from typing import List

import numpy as np
import torch

from achatbot.types.avatar import AudioSlice


class IFaceAvatar(ABC):
    """
    this is the interface for face avatar (now just for liteavatar method)
    - load model weights
    - audio2signal -> signal2img -> mouth2full -> audio+video
    """

    @abstractmethod
    def load(self):
        raise NotImplementedError("must be implemented in the child class")

    @abstractmethod
    def audio2signal(self, audio_slice: AudioSlice) -> list:
        raise NotImplementedError("must be implemented in the child class")

    @abstractmethod
    def signal2img(self, signal_data) -> tuple[torch.Tensor, int]:
        raise NotImplementedError("must be implemented in the child class")

    @abstractmethod
    def mouth2full(
        self, mouth_image: torch.Tensor, bg_frame_id: int, use_bg: bool = False
    ) -> np.ndarray:
        raise NotImplementedError("must be implemented in the child class")



import re
from typing import Dict

from achatbot.common.interface import ILlm
from achatbot.common.factory import EngineClass
from achatbot.common.chat_history import ChatHistory
from achatbot.common.utils.thread_safe import ThreadSafeDict


class BaseLLM(EngineClass):
    def __init__(self, **kwargs):
        super().__init__()
        self.session_chat_history: Dict[str, ChatHistory] = ThreadSafeDict()

    def get_session_chat_history(self, session_id: str) -> list:
        history = self.session_chat_history.get(session_id)
        return history.to_list() if history else []

    def model_name(self):
        if hasattr(self.args, "model_name"):
            return self.args.model_name
        return ""

    def _have_special_char(self, content: str, use_nltk: bool = False) -> int:
        """
        check content to match a specail char which is end of sentence
        !TODO: use NLTK sent_tokenize to get sentences. @weedge
        """
        if use_nltk is False:
            pattern = r"""[.。,，;；!！?？、]"""
            matches = re.findall(pattern, content)
            if len(matches) == 0:
                return -1
            return content.index(matches[len(matches) - 1])



import time
import wave

from modelscope.pipelines import pipeline
from modelscope.utils.constant import Tasks
from modelscope.fileio import File
from modelscope.outputs import OutputKeys
import typer
import numpy as np

from src.common.utils.audio_resample import (
    resample_bytes2bytes,
    resample_file,
    resample_bytes2numpy,
)
from src.modules.speech.help.audio_mock import generate_white_noise

app = typer.Typer()


"""
- https://modelscope.cn/models/iic/speech_dfsmn_ans_psm_48k_causal
"""


def create_wav_header(dataflow, sample_rate=16000, num_channels=1, bits_per_sample=16):
    """
    创建WAV文件头的字节串。

    :param dataflow: 音频bytes数据（以字节为单位）。
    :param sample_rate: 采样率，默认16000。
    :param num_channels: 声道数，默认1（单声道）。
    :param bits_per_sample: 每个样本的位数，默认16。
    :return: WAV文件头的字节串和音频bytes数据。
    """
    total_data_len = len(dataflow)
    byte_rate = sample_rate * num_channels * bits_per_sample // 8
    block_align = num_channels * bits_per_sample // 8
    data_chunk_size = total_data_len
    fmt_chunk_size = 16
    riff_chunk_size = 4 + (8 + fmt_chunk_size) + (8 + data_chunk_size)

    # 使用 bytearray 构建字节串
    header = bytearray()

    # RIFF/WAVE header
    header.extend(b"RIFF")
    header.extend(riff_chunk_size.to_bytes(4, byteorder="little"))
    header.extend(b"WAVE")

    # fmt subchunk
    header.extend(b"fmt ")
    header.extend(fmt_chunk_size.to_bytes(4, byteorder="little"))
    header.extend((1).to_bytes(2, byteorder="little"))  # Audio format (1 is PCM)
    header.extend(num_channels.to_bytes(2, byteorder="little"))
    header.extend(sample_rate.to_bytes(4, byteorder="little"))
    header.extend(byte_rate.to_bytes(4, byteorder="little"))
    header.extend(block_align.to_bytes(2, byteorder="little"))
    header.extend(bits_per_sample.to_bytes(2, byteorder="little"))

    # data subchunk
    header.extend(b"data")
    header.extend(data_chunk_size.to_bytes(4, byteorder="little"))

    return bytes(header) + dataflow


@app.command()
def online_stream():
    ans = pipeline(
        Tasks.acoustic_noise_suppression,
        model="iic/speech_dfsmn_ans_psm_48k_causal",  # output 48K auido sample rate
        stream_mode=True,
    )
    block_bytes_size = 160 * 2  # 模型每次处理的音频片段不能小于block_size

    start = time.time()
    white_noise = generate_white_noise(duration=0.1, sr=16000)
    audio_data = white_noise.astype(np.int16).tobytes()
    audio_48k_bytes = resample_bytes2bytes(
        audio_data,
        orig_freq=16000,
        new_freq=48000,
    )
    result = ans(audio_48k_bytes)
    out_pcm = result[OutputKeys.OUTPUT_PCM]
    print(f"warmup {len(out_pcm)=} cost: {time.time() - start} s")

    # audio_path = "./records/speech_with_noise_48k.wav"

    # input_path = "./records/speech_with_noise16k.wav"
    # audio_path = "./records/speech_with_noise48k.wav"
    # resample_file(input_path, audio_path)

    audio_path = "./records/speech_with_noise16k.wav"

    audio_pcm = b""
    begin = time.time()
    with open(os.path.join(os.getcwd(), audio_path), "rb") as f:
        # audio = f.read(block_bytes_size)
        audio = f.read(block_bytes_size + 44)[44:]  # remove RIFF header
        while len(audio) > 0:
            start = time.time()
            print("before", len(audio))
            audio_48k = resample_bytes2bytes(
                audio,
                orig_freq=16000,
                new_freq=48000,
            )
            print("resample after", len(audio_48k))
            result = ans(audio_48k)
            out_pcm = result[OutputKeys.OUTPUT_PCM]
            print("ans after", len(out_pcm))
            if len(out_pcm) > 0:
                out_pcm = resample_bytes2bytes(
                    out_pcm,
                    orig_freq=48000,
                    new_freq=16000,
                )
                print("resample back", len(out_pcm))
                audio_pcm += out_pcm
            print(f"cost: {time.time() - start} s")
            audio = f.read(block_bytes_size)

    print(f"total cost: {time.time() - begin} s")
    out_path = "./records/output_dfsmn_enhancer_stream.wav"
    with wave.open(out_path, "wb") as wf:
        wf.setnchannels(1)
        wf.setsampwidth(2)
        wf.setframerate(16000)
        wf.writeframes(audio_pcm)
        print(f"{audio_path=} save to {out_path}")


@app.command()
def offline_stream():
    ans = pipeline(Tasks.acoustic_noise_suppression, model="iic/speech_dfsmn_ans_psm_48k_causal")

    # audio_path = 'https://modelscope.cn/api/v1/models/damo/speech_dfsmn_ans_psm_48k_causal/repo?Revision=master&FilePath=examples/speech_with_noise_48k.wav',
    audio_path = "./records/speech_with_noise_48k.wav"

    if audio_path.startswith("http"):
        import io

        file_bytes = File.read(audio_path)
        audiostream = io.BytesIO(file_bytes)
    else:
        audiostream = open(audio_path, "rb")

    window = 1 * 48000 * 2  # 2 秒的窗口大小，以字节为单位
    outputs = b""
    total_bytes_len = 0
    audiostream.read(44)
    for dataflow in iter(lambda: audiostream.read(window), ""):
        print(len(dataflow))
        total_bytes_len += len(dataflow)
        if len(dataflow) == 0:
            break
        result = ans(
            create_wav_header(dataflow, sample_rate=48000, num_channels=1, bits_per_sample=16)
        )
        output = result["output_pcm"]
        outputs = outputs + output
    audiostream.close()

    outputs = outputs[:total_bytes_len]
    output_path = "dfsmn_enhancer_output_stream.wav"
    with open(output_path, "wb") as out_wave:
        out_wave.write(
            create_wav_header(outputs, sample_rate=48000, num_channels=1, bits_per_sample=16)
        )


@app.command()
def offline():
    ans = pipeline(Tasks.acoustic_noise_suppression, model="iic/speech_dfsmn_ans_psm_48k_causal")

    audio_path = "./records/speech_with_noise16k.wav"
    # audio_path = "./records/speech_with_noise_48k.wav"

    result = ans(
        # 'https://modelscope.cn/api/v1/models/damo/speech_dfsmn_ans_psm_48k_causal/repo?Revision=master&FilePath=examples/speech_with_noise_48k.wav',
        audio_path,
        output_path="dfsmn_enhancer_output.wav",
    )
    print(len(result["output_pcm"]))


"""
python -m demo.modelscope.ans_dfsmn_enhancer offline
python -m demo.modelscope.ans_dfsmn_enhancer offline-stream
python -m demo.modelscope.ans_dfsmn_enhancer online-stream
"""
if __name__ == "__main__":
    import torch
    import os

    thread_count = os.cpu_count()
    print(f"thread_count: {thread_count}")

    torch.set_num_threads(min(8, thread_count))
    torch.set_num_interop_threads(min(8, thread_count))

    app()



import os
import time
import subprocess
import asyncio

import modal

ASR_TAG = os.getenv("ASR_TAG", "whisper_transformers_torch_compile_asr")

app = modal.App(f"torch-compile-run-whisper-achatbot")

image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "cmake", "ninja-build")
    .pip_install("wheel")
    .pip_install(
        "accelerate",
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
        "torchcodec==0.2.0",
    )
    .pip_install("huggingface_hub[hf_transfer]", "transformers==4.52.4")
    .pip_install("librosa", "soundfile")
    .env(
        {
            "HF_HUB_ENABLE_HF_TRANSFER": "1",  # faster model transfers
            "TORCH_CUDA_ARCH_LIST": "8.0 8.9 9.0+PTX",
            "ASR_TAG": ASR_TAG,
            "TORCHDYNAMO_VERBOSE": "1",
            # "TORCH_LOGS": "+dynamo",
            "TORCHINDUCTOR_FX_GRAPH_CACHE": "1",  # https://docs.pytorch.org/tutorials/recipes/torch_compile_caching_tutorial.html
        }
    )
)

MODEL_DIR = "/root/.achatbot/models"
model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)

image = image.pip_install(
    f"achatbot=={os.getenv('ACHATBOT_VERSION', '0.0.24')}",
    extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    retries=0,
    image=image,
    volumes={
        MODEL_DIR: model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
)
async def run(
    func: callable = None,
    **kwargs,
) -> None:
    import torch

    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("which nvcc", shell=True)
    subprocess.run("nvcc --version", shell=True)
    print("cuda:", torch.version.cuda)
    print("_GLIBCXX_USE_CXX11_ABI", torch._C._GLIBCXX_USE_CXX11_ABI)

    print("torch:", torch.__version__)
    subprocess.run("pip show transformers", shell=True)
    subprocess.run("pip show achatbot", shell=True)

    print(f"{kwargs=}")
    if asyncio.iscoroutinefunction(func):
        await func(**kwargs)
    else:
        func(**kwargs)


async def transcribe(**kwargs):
    from achatbot.common.interface import IAsr
    from achatbot.modules.speech.asr import ASREnvInit
    from achatbot.common.session import Session
    from achatbot.common.types import SessionCtx
    from achatbot.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    asr: IAsr = ASREnvInit.initASREngine(
        os.getenv("ASR_TAG", "whisper_transformers_torch_compile_asr"), **kwargs
    )
    audio_file = os.path.join(ASSETS_DIR, "asr_example_zh.wav")
    session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
    asr.set_audio_data(audio_file)
    res = await asr.transcribe(session)
    print(res)


async def run_transcribe(**kwargs):
    import librosa
    import torch
    from torch.nn.attention import sdpa_kernel, SDPBackend
    from transformers import AutoModelForSpeechSeq2Seq, AutoProcessor

    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    torch_dtype = torch.float16 if torch.cuda.is_available() else torch.float32

    model_id = os.path.join(
        MODEL_DIR, kwargs.get("model_name_or_path", "openai/whisper-large-v3-turbo")
    )
    model = AutoModelForSpeechSeq2Seq.from_pretrained(
        model_id, torch_dtype=torch_dtype, low_cpu_mem_usage=True
    )
    model.to(device)

    processor = AutoProcessor.from_pretrained(model_id)

    inference_time = 0.0
    model.generation_config.max_new_tokens = 128
    torch.set_float32_matmul_precision("high")

    audio_file = os.path.join(ASSETS_DIR, "asr_example_zh.wav")
    audio_np, _ = librosa.load(audio_file, sr=16000)

    # 1. Pre-process the audio inputs
    # input_features: torch.Size([1, 80, 3000]) torch.float32 for v2 v3 support 128
    input_features = processor(audio_np, sampling_rate=16000, return_tensors="pt").input_features
    print(input_features.shape, input_features.dtype)
    input_features = input_features.to(device, dtype=torch_dtype)
    # Create an attention mask
    attention_mask = torch.ones(input_features.shape[:2], dtype=torch.long, device=device)
    # 2. Auto-regressively generate text tokens
    start = time.time()
    pred_ids = model.generate(input_features, attention_mask=attention_mask)
    inference_time = time.time() - start
    # 3. Post-process tokens to text
    pred_text = processor.batch_decode(pred_ids, skip_special_tokens=True)
    print(pred_text[0].lower())
    print(f"torch.compile Before {inference_time=}")

    # use torch compile
    model.forward = torch.compile(model.forward, mode="reduce-overhead", fullgraph=True)
    model.generation_config.cache_implementation = "static"
    max_new_tokens = model.generation_config.max_new_tokens
    for i in range(3):
        start = time.time()
        with sdpa_kernel(SDPBackend.MATH):
            model.generate(
                input_features,
                attention_mask=attention_mask,
                min_new_tokens=max_new_tokens,
                max_new_tokens=max_new_tokens,
            )
        inference_time = time.time() - start
        print(f"warmup {i} step {inference_time=}")

    # 1. Pre-process the audio inputs
    # input_features: torch.Size([1, 80, 3000]) torch.float32 for v2 v3 support 128
    inputs = processor(
        audio_np,
        sampling_rate=16000,
        return_tensors="pt",
        padding="max_length",
        return_attention_mask=True,
    )
    inputs = inputs.to(device, dtype=torch_dtype)
    print(inputs)
    print(inputs.input_features.shape)

    # Create an attention mask
    attention_mask = torch.ones(inputs.input_features.shape[:2], dtype=torch.long, device=device)
    print(attention_mask.shape, inputs.attention_mask.shape)

    def gen(inputs: dict, **kwargs):
        with sdpa_kernel(SDPBackend.MATH):
            pred_ids = model.generate(**inputs, **kwargs)
            return pred_ids

    # 2. Auto-regressively generate text tokens
    start = time.time()

    # with sdpa_kernel(SDPBackend.MATH):
    #    pred_ids = model.generate(
    #        **inputs,
    #        # inputs.input_features,
    #        # attention_mask=attention_mask,
    #        task="transcribe",
    #    )

    pred_ids = await asyncio.to_thread(gen, inputs, task="transcribe")
    print(pred_ids)
    inference_time = time.time() - start
    # 3. Post-process tokens to text
    pred_text = processor.batch_decode(
        pred_ids, skip_special_tokens=True, decode_with_timestamps=True
    )
    print(pred_text[0].lower())
    print(f"torch.compile After {inference_time=}")


async def run_pipeline_transcribe(**kwargs):
    import librosa
    import torch
    from torch.nn.attention import sdpa_kernel, SDPBackend
    from transformers import AutoModelForSpeechSeq2Seq, AutoProcessor, pipeline

    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    torch_dtype = torch.float16 if torch.cuda.is_available() else torch.float32

    model_id = os.path.join(
        MODEL_DIR, kwargs.get("model_name_or_path", "openai/whisper-large-v3-turbo")
    )
    model = AutoModelForSpeechSeq2Seq.from_pretrained(
        model_id, torch_dtype=torch_dtype, low_cpu_mem_usage=True
    )
    model.to(device)

    processor = AutoProcessor.from_pretrained(model_id)
    pipe = pipeline(
        "automatic-speech-recognition",
        model=model,
        tokenizer=processor.tokenizer,
        feature_extractor=processor.feature_extractor,
        chunk_length_s=30,
        batch_size=1,  # batch size for inference - set based on your device
        torch_dtype=torch_dtype,
        device=device,
    )

    audio_file = os.path.join(ASSETS_DIR, "asr_example_zh.wav")
    audio_np, _ = librosa.load(audio_file, sr=16000)

    start = time.time()
    result = pipe(
        audio_np.copy(),
        chunk_length_s=30,
        batch_size=1,
        # generate_kwargs={"language": "zh", "task": "transcribe"},
        return_timestamps=True,
        # return_timestamps="word",
    )
    inference_time = time.time() - start
    print(result)
    print(f"torch.compile Before {inference_time=}")

    # 注意： torch.compile 目前与 Chunked 长格式算法或 Flash Attention 2 不兼容⚠️
    # Enable static cache and compile the forward pass
    model.generation_config.cache_implementation = "static"
    model.generation_config.max_new_tokens = 256
    model.forward = torch.compile(model.forward, mode="reduce-overhead", fullgraph=True)
    pipe = pipeline(
        "automatic-speech-recognition",
        model=model,
        tokenizer=processor.tokenizer,
        feature_extractor=processor.feature_extractor,
        torch_dtype=torch_dtype,
        device=device,
    )

    # 2 warmup steps
    for i in range(2):
        start = time.time()
        with sdpa_kernel(SDPBackend.MATH):
            result = pipe(
                audio_np.copy(),
                generate_kwargs={"min_new_tokens": 256, "max_new_tokens": 256},
                return_timestamps=True,
            )
        inference_time = time.time() - start
        print(f"warmup {i} step {inference_time=}")

    def pipe_gen(audio_np, **kwargs):
        with sdpa_kernel(SDPBackend.MATH):
            return pipe(audio_np, **kwargs)

    # fast run
    start = time.time()
    # with sdpa_kernel(SDPBackend.MATH):
    #    result = pipe(
    #        audio_np.copy(),
    #        chunk_length_s=30,
    #        batch_size=1,
    #        return_timestamps=True,
    #        # return_timestamps="word",
    #    )

    result = await asyncio.to_thread(
        pipe_gen,
        audio_np,
        generate_kwargs={
            "language": "zh",
            "task": "transcribe",
        },
        chunk_length_s=30,
        batch_size=1,
        return_timestamps=True,
        # return_timestamps="word",
    )
    inference_time = time.time() - start
    print(f"{inference_time=}")
    print(len(result["text"]), len(result["chunks"]))
    print(result["chunks"])


"""
https://huggingface.co/openai/whisper-large-v3
https://huggingface.co/openai/whisper-large-v3-turbo

# download model weights
modal run src/download_models.py --repo-ids "openai/whisper-large-v3"
modal run src/download_models.py --repo-ids "openai/whisper-large-v3-turbo"

ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3-turbo"
IMAGE_GPU=A100 ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3-turbo"

ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3"
IMAGE_GPU=A100 ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3"

ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_pipeline_transcribe --model-name-or-path "openai/whisper-large-v3-turbo"
IMAGE_GPU=A100 ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3-turbo"

ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_pipeline_transcribe --model-name-or-path "openai/whisper-large-v3"
IMAGE_GPU=A100 ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task run_transcribe --model-name-or-path "openai/whisper-large-v3"

# use large-v3-turbo model with torch compile
ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task transcribe --model-name-or-path "openai/whisper-large-v3-turbo"
ASR_TAG=whisper_transformers_pipeline_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task transcribe --model-name-or-path "openai/whisper-large-v3-turbo"

# use large-v3 model with torch compile
ASR_TAG=whisper_transformers_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task transcribe --model-name-or-path "openai/whisper-large-v3"
ASR_TAG=whisper_transformers_pipeline_torch_compile_asr modal run src/llm/achatbot_torch_compile_audio.py --task transcribe --model-name-or-path "openai/whisper-large-v3"

"""


@app.local_entrypoint()
def main(
    task: str = "transcribe",
    model_name_or_path: str = "openai/whisper-large-v3-turbo",
    # https://docs.pytorch.org/docs/stable/generated/torch.compile.html
    torch_compile_mode: str = "reduce-overhead",
):
    tasks = {
        "transcribe": transcribe,
        "run_transcribe": run_transcribe,
        "run_pipeline_transcribe": run_pipeline_transcribe,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    model_name_or_path = os.path.join(MODEL_DIR, model_name_or_path)
    run.remote(
        func=tasks[task],
        model_name_or_path=model_name_or_path,
        torch_compile_mode=torch_compile_mode,
    )



from dataclasses import dataclass
import io
import os
import logging
from typing import AsyncGenerator

import aiohttp
from PIL import Image
from apipeline.frames.data_frames import Frame, ImageRawFrame
from apipeline.frames.sys_frames import ErrorFrame

from achatbot.processors.image.base import ImageGenProcessor


# https://huggingface.co/docs/api-inference/tasks/text-to-image


class HFApiInferenceImageGenProcessor(ImageGenProcessor):
    def __init__(
        self,
        *,
        aiohttp_session: aiohttp.ClientSession = None,
        model: str = "stabilityai/stable-diffusion-3.5-large",
        width: int = 1024,
        height: int = 1024,
        steps: int = 28,
    ):
        super().__init__()
        self._model = model
        self._width = width
        self._height = height
        self._steps = steps
        self._api_key = os.environ.get("HF_API_KEY")
        self._aiohttp_session = aiohttp_session

    async def run_image_gen(self, prompt: str) -> AsyncGenerator[Frame, None]:
        logging.debug(f"Generating image from prompt: {prompt}")

        API_URL = f"https://api-inference.huggingface.co/models/{self._model}"
        headers = {"Authorization": f"Bearer {self._api_key}"}
        response = await self._aiohttp_session.post(
            API_URL,
            headers=headers,
            json={
                "inputs": prompt,
                # target_size is not work, use default params
                # "target_size": {
                #    "width": self._width,
                #    "height": self._height,
                # },
                # "num_inference_steps": self._steps,
            },
        )
        image_bytes = await response.content.read()

        if not image_bytes:
            yield ErrorFrame("Image generation failed")
            return

        image_stream = io.BytesIO(image_bytes)
        image = Image.open(image_stream).convert("RGB")
        # if image.size != (self._width, self._height):
        #    image = image.resize((self._width, self._height))
        frame = ImageRawFrame(
            image=image.tobytes(),
            size=image.size,
            format=image.format if image.format else "JPEG",
            mode=image.mode,
        )
        yield frame



import asyncio
import logging
import string
from abc import abstractmethod
from typing import AsyncGenerator

from apipeline.processors.frame_processor import FrameDirection, FrameProcessorMetrics
from apipeline.frames.control_frames import EndFrame, StartFrame
from apipeline.frames.sys_frames import Frame, MetricsFrame, StartInterruptionFrame, CancelFrame
from apipeline.utils.string import match_endofsentence

from src.processors.ai_processor import AIProcessor
from src.types.frames.control_frames import (
    LLMFullResponseEndFrame,
    TTSStartedFrame,
    TTSStoppedFrame,
    TTSVoiceUpdateFrame,
)
from src.types.frames.data_frames import TTSSpeakFrame, TextFrame


class TTSProcessorMetrics(FrameProcessorMetrics):
    async def start_tts_usage_metrics(self, text: str):
        characters = {
            "processor": self._name,
            "value": len(text),
        }
        logging.debug(f"{self._name} usage characters: {characters['value']}")
        return MetricsFrame(characters=[characters])


class TTSProcessorBase(AIProcessor):
    def __init__(
        self,
        *,
        aggregate_sentences: bool = True,
        # if True, subclass is responsible for pushing TextFrames and LLMFullResponseEndFrames
        push_text_frames: bool = True,
        sync_order_send: bool = False,
        remove_punctuation: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._metrics = TTSProcessorMetrics(name=self.name)

        self._aggregate_sentences: bool = aggregate_sentences
        self._push_text_frames: bool = push_text_frames
        self._remove_punctuation: bool = remove_punctuation
        self._current_sentence: str = ""

        # sync event: tts done, step by step slow,
        # e.g. for test, push EndFrame in the end.
        self._tts_done_event = asyncio.Event()
        self._sync_order_send = sync_order_send

    def get_stream_info(self) -> dict:
        return {
            "sample_rate": 16000,
            "sample_width": 2,
            "channels": 1,
        }

    @abstractmethod
    async def set_voice(self, voice: str):
        pass

    @abstractmethod
    async def set_tts_args(self, **args):
        pass

    @abstractmethod
    async def run_tts(self, text: str) -> AsyncGenerator[Frame, None]:
        pass

    async def start(self, frame: StartFrame):
        logging.info(f"{self.name} started")

    async def stop(self, frame: EndFrame):
        logging.info(f"{self.name} end")

    async def cancel(self, frame: CancelFrame):
        logging.info(f"{self.name} canceled")

    async def start_tts_usage_metrics(self, text: str):
        if self.can_generate_metrics() and self.usage_metrics_enabled:
            frame = await self._metrics.start_tts_usage_metrics(text)
            if frame:
                await self.push_frame(frame)

    async def say(self, text: str):
        await self.process_frame(TextFrame(text=text), FrameDirection.DOWNSTREAM)

    async def _handle_interruption(self, frame: StartInterruptionFrame, direction: FrameDirection):
        self._current_sentence = ""
        await self.push_frame(frame, direction)

    async def _process_text_frame(self, frame: TextFrame):
        text: str | None = None
        if not self._aggregate_sentences:
            text = frame.text
        else:
            self._current_sentence += frame.text
            if match_endofsentence(self._current_sentence):
                text = self._current_sentence
                self._current_sentence = ""
        # print(f"frame: {frame} text: {text}")
        if text:
            await self._push_tts_frames(text)

    async def _push_tts_frames(self, text: str, text_passthrough: bool = True):
        text = text.strip()
        if not text:
            return

        # show text sentence before tts speak
        if self._push_text_frames:
            # We send the original text after the audio. This way, if we are
            # interrupted, the text is not added to the assistant context.
            await self.push_frame(TextFrame(text))

        await self.push_frame(TTSStartedFrame())
        await self.start_processing_metrics()

        if self._remove_punctuation:
            translator = str.maketrans("", "", string.punctuation)
            text = text.translate(translator)
        await self.process_generator(self.run_tts(text))
        # !NOTE: when open sync order send frame, u need set sync order
        # run_tts send audio frames over then send tts stopped frame
        # need subclass _tts_done_event to set;
        if self._sync_order_send is True:
            await self._tts_done_event.wait()
            self._tts_done_event.clear()

        await self.stop_processing_metrics()
        await self.push_frame(TTSStoppedFrame())

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, TextFrame):
            await self._process_text_frame(frame)
        elif isinstance(frame, StartInterruptionFrame):
            await self._handle_interruption(frame, direction)
        elif isinstance(frame, LLMFullResponseEndFrame) or isinstance(frame, EndFrame):
            sentence = self._current_sentence
            self._current_sentence = ""
            await self._push_tts_frames(sentence)
            if isinstance(frame, LLMFullResponseEndFrame):
                if self._push_text_frames:
                    await self.push_frame(frame, direction)
            else:
                await self.push_frame(frame, direction)
        elif isinstance(frame, TTSSpeakFrame):
            await self._push_tts_frames(frame.text, False)
        elif isinstance(frame, TTSVoiceUpdateFrame):
            await self.set_voice(frame.voice)
        else:
            await self.push_frame(frame, direction)



import io
import wave
import logging

from fastapi import WebSocket
from fastapi.websockets import WebSocketState
from apipeline.frames.data_frames import Frame, AudioRawFrame, TextFrame
from apipeline.frames.sys_frames import StartInterruptionFrame
from apipeline.processors.frame_processor import FrameDirection

from src.processors.audio_camera_output_processor import AudioCameraOutputProcessor
from src.types.network.fastapi_websocket import FastapiWebsocketServerParams
from src.types.frames.data_frames import AnimationAudioRawFrame


class FastapiWebsocketServerOutputProcessor(AudioCameraOutputProcessor):
    def __init__(self, websocket: WebSocket, params: FastapiWebsocketServerParams, **kwargs):
        super().__init__(params, **kwargs)

        self._websocket = websocket
        self._params = params
        self.audio_frame_size = (
            self._params.audio_out_channels
            * self._params.audio_out_sample_width
            * (self._params.audio_out_sample_rate * self._params.audio_out_frame_ms // 1000)
        )
        self._websocket_audio_buffer = bytes()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartInterruptionFrame):
            await self._write_frame(frame)

    async def send_text(self, frame: TextFrame):
        await self.send_payload(frame)

    async def write_raw_audio_frames(self, frames: bytes):
        self._websocket_audio_buffer += frames
        while len(self._websocket_audio_buffer):
            frame = AudioRawFrame(
                audio=self._websocket_audio_buffer[: self.audio_frame_size],
                sample_rate=self._params.audio_out_sample_rate,
                num_channels=self._params.audio_out_channels,
            )

            if self._params.add_wav_header:
                content = io.BytesIO()
                ww = wave.open(content, "wb")
                ww.setsampwidth(frame.sample_width)
                ww.setnchannels(frame.num_channels)
                ww.setframerate(frame.sample_rate)
                ww.writeframes(frame.audio)
                ww.close()
                content.seek(0)
                wav_frame = AudioRawFrame(
                    content.read(), sample_rate=frame.sample_rate, num_channels=frame.num_channels
                )
                frame = wav_frame

            await self.send_payload(frame)

            self._websocket_audio_buffer = self._websocket_audio_buffer[self.audio_frame_size :]

    async def write_animation_audio_frame(self, frame: AnimationAudioRawFrame):
        if self._params.add_wav_header and frame.audio:
            content = io.BytesIO()
            ww = wave.open(content, "wb")
            ww.setsampwidth(frame.sample_width)
            ww.setnchannels(frame.num_channels)
            ww.setframerate(frame.sample_rate)
            ww.writeframes(frame.audio)
            ww.close()
            content.seek(0)
            frame.audio = content.read()

        await self.send_payload(frame)

    async def _write_frame(self, frame: Frame):
        await self.send_payload(frame)

    async def send_payload(self, frame: Frame):
        try:
            payload = self._params.serializer.serialize(frame)
            if not payload:
                logging.warning(f"serialize frame: {frame} no payload")
                return
            if self._websocket.client_state != WebSocketState.CONNECTED:
                logging.warning(
                    f"websocket not connected, client_state:{self._websocket.client_state}"
                )
                return

            if isinstance(payload, str):
                await self._websocket.send_text(payload)
            else:
                await self._websocket.send_bytes(payload)
            # logging.debug(f"send payload: {type(payload)} len:{len(payload)}")
        except Exception as e:
            logging.exception(f"send_payload error: {e}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tts.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ttts.proto\x12\x0c\x63hat_bot.tts\"K\n\x10LoadModelRequest\x12\x0f\n\x07tts_tag\x18\x01 \x01(\t\x12\x11\n\tis_reload\x18\x02 \x01(\x08\x12\x13\n\x0bjson_kwargs\x18\x03 \x01(\t\"\x13\n\x11LoadModelResponse\":\n\x11SynthesizeRequest\x12\x10\n\x08tts_text\x18\x01 \x01(\t\x12\x13\n\x0bjson_kwargs\x18\x02 \x01(\t\"\'\n\x12SynthesizeResponse\x12\x11\n\ttts_audio\x18\x01 \x01(\x0c\"\x12\n\x10GetVoicesRequest\"#\n\x11GetVoicesResponse\x12\x0e\n\x06voices\x18\x01 \x03(\t\" \n\x0fSetVoiceRequest\x12\r\n\x05voice\x18\x01 \x01(\t\"\x12\n\x10SetVoiceResponse\"\x16\n\x14GetStreamInfoRequest\"\\\n\x14GetStreamInfoReponse\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\x05\x12\x10\n\x08\x63hannels\x18\x02 \x01(\x05\x12\x0c\n\x04rate\x18\x03 \x01(\x05\x12\x14\n\x0csample_width\x18\x04 \x01(\x05\x32\xa4\x03\n\x03TTS\x12N\n\tLoadModel\x12\x1e.chat_bot.tts.LoadModelRequest\x1a\x1f.chat_bot.tts.LoadModelResponse\"\x00\x12N\n\tGetVoices\x12\x1e.chat_bot.tts.GetVoicesRequest\x1a\x1f.chat_bot.tts.GetVoicesResponse\"\x00\x12K\n\x08SetVoice\x12\x1d.chat_bot.tts.SetVoiceRequest\x1a\x1e.chat_bot.tts.SetVoiceResponse\"\x00\x12Y\n\rGetStreamInfo\x12\".chat_bot.tts.GetStreamInfoRequest\x1a\".chat_bot.tts.GetStreamInfoReponse\"\x00\x12U\n\x0cSynthesizeUS\x12\x1f.chat_bot.tts.SynthesizeRequest\x1a .chat_bot.tts.SynthesizeResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tts_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_LOADMODELREQUEST']._serialized_start=27
  _globals['_LOADMODELREQUEST']._serialized_end=102
  _globals['_LOADMODELRESPONSE']._serialized_start=104
  _globals['_LOADMODELRESPONSE']._serialized_end=123
  _globals['_SYNTHESIZEREQUEST']._serialized_start=125
  _globals['_SYNTHESIZEREQUEST']._serialized_end=183
  _globals['_SYNTHESIZERESPONSE']._serialized_start=185
  _globals['_SYNTHESIZERESPONSE']._serialized_end=224
  _globals['_GETVOICESREQUEST']._serialized_start=226
  _globals['_GETVOICESREQUEST']._serialized_end=244
  _globals['_GETVOICESRESPONSE']._serialized_start=246
  _globals['_GETVOICESRESPONSE']._serialized_end=281
  _globals['_SETVOICEREQUEST']._serialized_start=283
  _globals['_SETVOICEREQUEST']._serialized_end=315
  _globals['_SETVOICERESPONSE']._serialized_start=317
  _globals['_SETVOICERESPONSE']._serialized_end=335
  _globals['_GETSTREAMINFOREQUEST']._serialized_start=337
  _globals['_GETSTREAMINFOREQUEST']._serialized_end=359
  _globals['_GETSTREAMINFOREPONSE']._serialized_start=361
  _globals['_GETSTREAMINFOREPONSE']._serialized_end=453
  _globals['_TTS']._serialized_start=456
  _globals['_TTS']._serialized_end=876
# @@protoc_insertion_point(module_scope)



class ChatHistory:
    """
    buffer the local chat hostory with limit size using to avoid llm context too long.
    - if size is None, no limit
    - if size < 0, no history

    !TODO: use kv store history like mem0. @weedge
    """

    def __init__(
        self, size: int | None = None, init_chat_message: dict = None, init_chat_tools: dict = None
    ):
        self.size = size
        self.init_chat_message = init_chat_message
        self.init_chat_tools = init_chat_tools
        # maxlen is necessary pair,
        # since a each new step we add an prompt and assitant answer
        self.buffer = []

    def set_size(self, size: int | None):
        self.size = size

    def clear(self):
        self.buffer.clear()

    def append(self, item):
        if self.size and self.size < 0:
            return

        self.buffer.append(item)
        if self.size is None:
            return

        if len(self.buffer) == 2 * (self.size + 1):
            self.buffer.pop(0)
            self.buffer.pop(0)

    def pop(self, index: int = -1):
        if self.size and self.size < 0:
            return
        if len(self.buffer) > 0:
            self.buffer.pop(index)

    def init(self, init_chat_message: dict):
        self.init_chat_message = init_chat_message

    def init_tools(self, tools: dict):
        self.init_chat_tools = tools

    def to_list(self) -> list:
        if self.init_chat_message:
            if self.init_chat_tools:
                return [self.init_chat_message, self.init_chat_tools] + self.buffer
            else:
                return [self.init_chat_message] + self.buffer
        else:
            return self.buffer

    def __getstate__(self):
        return {
            "size": self.size,
            "init_chat_message": self.init_chat_message,
            "init_chat_tools": self.init_chat_tools,
            "buffer": self.buffer,
        }

    def __setstate__(self, state):
        self.size = state["size"]
        self.init_chat_message = state["init_chat_message"]
        self.init_chat_tools = state["init_chat_tools"]
        self.buffer = state["buffer"]

    def __repr__(self) -> str:
        chat_history = self.__getstate__()
        return f"{chat_history}"


"""
python src/common/chat_history.py
"""
if __name__ == "__main__":
    chat_history = ChatHistory(size=2)
    print(chat_history)
    chat_history.append({"role": "user", "content": "Hello 0"})
    chat_history.append({"role": "assistant", "content": "Hi, how can I help you 0?"})
    print(chat_history)
    chat_history.append({"role": "user", "content": "Hello 1"})
    chat_history.append({"role": "assistant", "content": "Hi, how can I help you 1?"})
    print(chat_history)
    chat_history.pop(-1)
    chat_history.append({"role": "assistant", "content": "Hi, how can I help you 1.1?"})

    chat_history.append({"role": "user", "content": "Hello 2"})
    chat_history.append({"role": "assistant", "content": "Hi, how can I help you 2?"})
    print(chat_history)

    chat_history.append({"role": "user", "content": "Hello 3"})
    chat_history.append({"role": "assistant", "content": "Hi, how can I help you 3?"})
    print(chat_history)



import logging
from threading import Thread
from time import perf_counter
from typing import Generator, Optional, Callable

import torch


try:
    from transformers import (
        Qwen2_5OmniForConditionalGeneration,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Qwen2.5Omni, you need to `pip install git+https://github.com/huggingface/transformers`"
    )
    raise Exception(f"Missing module: {e}")


from src.common.utils.helper import print_model_params
from src.core.llm.transformers.streamer import TokenStreamer


class Qwen2_5OmniForConditionalGenerationStreaming(Qwen2_5OmniForConditionalGeneration):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print_model_params(self.thinker, "qwen2.5omni_thinker")
        if self.has_talker:
            print_model_params(self.talker, "qwen2.5omni_talker")
            print_model_params(self.token2wav, "qwen2.5omni_token2wav")
            self.token2wav.float()

    @torch.no_grad()
    def generate(
        self,
        input_ids: Optional[torch.Tensor] = None,
        speaker: str = "Chelsie",
        use_audio_in_video: bool = False,
        return_audio: Optional[bool] = None,
        thinker_max_new_tokens: int = 1024,
        talker_max_new_tokens: int = 4096,
        talker_do_sample: bool = True,
        talker_top_k: int = 40,
        talker_top_p: float = 0.8,
        talker_temperature: float = 0.9,
        talker_eos_token_ids: list[int] = [8292, 8294],
        talker_repetition_penalty: float = 1.05,
        **kwargs,
    ):
        r"""
        Generate text response and audio from input.

        Args:
            input_ids (`Optional[torch.Tensor]`, *optional*):
                Input ids, should obtain from self._tokenizer.
            speaker (`str` , defaults to "Chelsie"):
                Which speaker should be used in audio response.
            use_audio_in_video (`bool`, defaults to False):
                Whether or not use audio track in video, should same as the parameter in `process_audio_info`.
            return_audio (`Optional[bool]`, *optional*):
                Whether or not return response in audio format. When `return_audio=None`, this parameter is same as `config.enable_audio_output`.
            kwargs (*optional*):
                - Without a prefix, they will be entered as `**kwargs` for the `generate` method of each sub-self.
                - With a *thinker_*, *talker_*, *token2wav_* prefix, they will be input for the `generate` method of the
                thinker, talker and token2wav respectively. It has the priority over the keywords without a prefix.
        Returns:
            When `return_audio=False`:
                - **Text** (`torch.Tensor`): Generated text token sequence.
            When `return_audio=True`:
                - **Text** (`torch.Tensor`): Generated text token sequence.
                - **Audio waveform** (`torch.Tensor`): Generated audio waveform.
        """
        if speaker not in self.speaker_map:
            raise ValueError(
                f"{speaker} is not availible, availible speakers: {self.speaker_map.keys()}"
            )
        if return_audio and not self.has_talker:
            raise ValueError(
                "Cannot use talker when talker module not initalized. Use `enable_talker` method or set enable_talker in config to enable talker."
            )
        if return_audio is None:
            return_audio = self.has_talker
        if input_ids.shape[0] != 1 and return_audio:
            raise NotImplementedError(
                "Qwen2.5-Omni currently does not support batched inference with audio output"
            )

        shared_kwargs = {"use_audio_in_video": use_audio_in_video}
        thinker_kwargs = {
            "max_new_tokens": thinker_max_new_tokens,
        }
        talker_kwargs = {
            "max_new_tokens": talker_max_new_tokens,
            "do_sample": talker_do_sample,
            "top_k": talker_top_k,
            "top_p": talker_top_p,
            "temperature": talker_temperature,
            "eos_token_id": talker_eos_token_ids,
            "repetition_penalty": talker_repetition_penalty,
        }
        token2wav_kwargs = {}

        for key, value in kwargs.items():
            if key.startswith("thinker_"):
                thinker_kwargs[key[len("thinker_") :]] = value
            elif key.startswith("talker_"):
                talker_kwargs[key[len("talker_") :]] = value
            elif key.startswith("token2wav_"):
                token2wav_kwargs[key[len("token2wav_") :]] = value
            # Process special input values
            elif key == "feature_attention_mask":
                thinker_kwargs[key] = value
                talker_kwargs["audio_feature_lengths"] = torch.sum(value, dim=1)
            elif key == "input_features" or key == "attention_mask":
                thinker_kwargs[key] = value
            # Put other key to shared kwargs
            else:
                shared_kwargs[key] = value

        # Merge kwargs
        for key, value in shared_kwargs.items():
            if key not in thinker_kwargs:
                thinker_kwargs[key] = value
            if key not in talker_kwargs:
                talker_kwargs[key] = value
            if key not in token2wav_kwargs:
                token2wav_kwargs[key] = value
        speaker_params = self.speaker_map[speaker]

        # 1. Generate from thinker module
        generate_audio = return_audio and self.has_talker
        if generate_audio:
            thinker_kwargs["output_hidden_states"] = True
            thinker_kwargs["return_dict_in_generate"] = True

        thinker_result = self.thinker.generate(input_ids=input_ids, **thinker_kwargs)

        if not generate_audio:
            return thinker_result

        # 2. Generate speech tokens from talker module
        embeds_to_talker = thinker_result.hidden_states[0][0].clone().to(self.talker.device)
        if thinker_kwargs.get("input_features", None) is not None:
            audio_ids_mask = input_ids == self.config.thinker_config.audio_token_index
            audio_mask = (
                audio_ids_mask.unsqueeze(-1).expand_as(embeds_to_talker).to(embeds_to_talker.device)
            )
            audio_mask_tensor = torch.zeros(
                [audio_ids_mask.sum(), embeds_to_talker.shape[-1]],
                dtype=embeds_to_talker.dtype,
                device=self.talker.device,
            )
            embeds_to_talker.masked_scatter_(audio_mask, audio_mask_tensor)
        if thinker_kwargs.get("pixel_values", None) is not None:
            image_ids_mask = input_ids == self.config.thinker_config.image_token_index
            image_mask = (
                image_ids_mask.unsqueeze(-1).expand_as(embeds_to_talker).to(embeds_to_talker.device)
            )
            image_mask_tensor = torch.zeros(
                [image_ids_mask.sum(), embeds_to_talker.shape[-1]],
                dtype=embeds_to_talker.dtype,
                device=self.talker.device,
            )
            embeds_to_talker.masked_scatter_(image_mask, image_mask_tensor)
        if thinker_kwargs.get("pixel_values_videos", None) is not None:
            video_ids_mask = input_ids == self.config.thinker_config.video_token_index
            video_mask = (
                video_ids_mask.unsqueeze(-1).expand_as(embeds_to_talker).to(embeds_to_talker.device)
            )
            video_mask_tensor = torch.zeros(
                [video_ids_mask.sum(), embeds_to_talker.shape[-1]],
                dtype=embeds_to_talker.dtype,
                device=self.talker.device,
            )
            embeds_to_talker.masked_scatter_(video_mask, video_mask_tensor)

        processed_thinker_hidden = (
            (embeds_to_talker,) + thinker_result.hidden_states[0][1:],
        ) + thinker_result.hidden_states[1:]

        thinker_generate_ids = thinker_result.sequences[:, input_ids.size(1) :].to(
            self.talker.device
        )
        thinker_token_embeds = [
            token_hidden_states[0].to(self.talker.device)
            for token_hidden_states in processed_thinker_hidden
        ]
        thinker_hidden_states = [
            token_hidden_states[-1].to(self.talker.device)
            for token_hidden_states in processed_thinker_hidden
        ]

        talker_text_bos_token = speaker_params["bos_token"]
        talker_input_text_ids = torch.cat(
            [
                input_ids.to(self.talker.device),
                torch.tensor(
                    [[talker_text_bos_token]], dtype=torch.long, device=self.talker.device
                ),
                thinker_generate_ids[:, :1],
            ],
            dim=-1,
        )

        talker_input_ids = torch.cat(
            [
                torch.full_like(
                    input_ids, fill_value=self.talker.codec_mask_token, device=self.talker.device
                ),
                torch.tensor(
                    [[self.talker.codec_pad_token]], dtype=torch.long, device=self.talker.device
                ),
                torch.tensor(
                    [[self.talker.codec_bos_token]], dtype=torch.long, device=self.talker.device
                ),
            ],
            dim=1,
        )

        thinker_embed_tokens = self.thinker.get_input_embeddings()
        thinker_reply_part = torch.cat(thinker_hidden_states[1:], dim=1) + torch.cat(
            thinker_token_embeds[1:], dim=1
        )
        talker_inputs_embeds = thinker_hidden_states[0] + thinker_token_embeds[0]
        talker_text_bos_token = torch.tensor(
            [[talker_text_bos_token]], dtype=torch.long, device=self.thinker.device
        )
        talker_text_bos_embed = thinker_embed_tokens(talker_text_bos_token).to(self.talker.device)
        talker_inputs_embeds = torch.cat(
            [
                talker_inputs_embeds,
                talker_text_bos_embed,
                thinker_reply_part[:, :1, :],
            ],
            dim=1,
        )

        eos_embedding = thinker_embed_tokens(
            torch.tensor(
                [[self.talker.text_eos_token]], dtype=torch.long, device=self.thinker.device
            )
        ).to(self.talker.device)

        pad_embedding = thinker_embed_tokens(
            torch.tensor(
                [[self.talker.text_pad_token]], dtype=torch.long, device=self.thinker.device
            )
        ).to(self.talker.device)

        thinker_reply_part = torch.cat(
            [
                thinker_reply_part[:, 1:, :],
                eos_embedding,
                pad_embedding,
            ],
            dim=1,
        )
        talker_attention_mask = None
        if "attention_mask" in kwargs:
            talker_attention_mask = torch.cat(
                [kwargs["attention_mask"], kwargs["attention_mask"].new_ones((1, 2))], dim=1
            ).to(self.talker.device)

        # stream
        skip_prompt = kwargs.get("skip_prompt", True)
        streamer = TokenStreamer(skip_prompt=skip_prompt)
        talker_kwargs = dict(
            input_ids=talker_input_ids,
            streamer=streamer,
            input_text_ids=talker_input_text_ids,
            thinker_reply_part=thinker_reply_part,
            inputs_embeds=talker_inputs_embeds,
            attention_mask=talker_attention_mask,
            suppress_tokens=[self.talker.codec_bos_token],
            **{
                k: (v.to(self.talker.device) if torch.is_tensor(v) else v)
                for k, v in talker_kwargs.items()
            },
        )
        # logging.debug(talker_kwargs.keys())
        thread = Thread(target=self.talker.generate, kwargs=talker_kwargs)
        thread.start()
        talker_generate_codes = []
        times = []
        start_time = perf_counter()
        for token_id in streamer:
            times.append(perf_counter() - start_time)
            start_time = perf_counter()
            talker_generate_codes.append(token_id)
        logging.info(
            f"generate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
        )
        offset = 0
        if skip_prompt is False:
            offset = talker_input_ids.shape[1]
        # logging.debug(
        #    talker_input_ids.shape[1],
        #    # talker_generate_codes,
        #    talker_generate_codes[:offset],
        #    talker_generate_codes[offset:-1],
        # )
        talker_generate_codes = torch.tensor(
            [talker_generate_codes[offset:-1]],
            dtype=torch.long,
            device=self.talker.device,
        )

        # logging.debug(f"talker_generate_codes:{talker_generate_codes.shape} {talker_generate_codes}")

        # 3. Generate wavs from code
        # logging.debug(self.token2wav.device, speaker_params, token2wav_kwargs)
        wav = self.token2wav(
            talker_generate_codes.to(self.token2wav.device),
            conditioning=speaker_params["cond"].to(self.token2wav.device).float(),
            reference_mel=speaker_params["ref_mel"].to(self.token2wav.device).float(),
            **token2wav_kwargs,
        )

        return thinker_result.sequences, wav.float()

    @torch.no_grad()
    def thinker_generate_chunk(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
        thinker_max_new_tokens: int = 1024,
        thinker_top_k: int = 40,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_eos_token_ids=[151643, 151645],  # Define EOS tokens
        thinker_repetition_penalty: float = 1.05,
        thinker_output_hidden_states=False,
        thinker_stop_strings_per_step=[],
        tokenizer=None,
        **kwargs,
    ):
        input_ids = inputs.pop("input_ids")
        attention_mask = inputs.pop("attention_mask", None)

        if thinker_max_tokens_per_step > thinker_max_new_tokens:
            thinker_max_tokens_per_step = thinker_max_new_tokens

        # Keep track of the full generated sequence full_generated_ids = input_ids.clone()
        # Ensure full_attention_mask is correctly initialized and expanded
        full_attention_mask = (
            attention_mask.clone()
            if attention_mask is not None
            else torch.ones_like(input_ids, device=input_ids.device)
        )
        full_generated_ids = input_ids.clone()

        # KV cache
        # past_key_values = None

        # Inputs for the current step
        current_input_ids = full_generated_ids
        # The attention mask passed to generate should cover the sequence length for the current step
        current_attention_mask = full_attention_mask

        total_new_tokens_generated = 0
        hidden_states = None
        hidden_states_len = 0

        times = []
        while total_new_tokens_generated < thinker_max_new_tokens:
            # Prepare inputs for generate call
            # logging.debug(current_input_ids, current_attention_mask.shape)
            model_inputs = {
                "input_ids": current_input_ids,
                "attention_mask": current_attention_mask,
                # "past_key_values": past_key_values,
                "use_cache": True,
                "use_audio_in_video": use_audio_in_video,
                "do_sample": True if thinker_temperature > 0 else False,
                "top_k": thinker_top_k,
                "top_p": thinker_top_p,
                "temperature": thinker_temperature,
                "repetition_penalty": thinker_repetition_penalty,
                "min_new_tokens": 1,  # Ensure at least one token is generated if possible
                "max_new_tokens": thinker_max_tokens_per_step,  # Generate in smaller steps
                # output_hidden_states/scores can consume memory,
                # enable if needed downstream(talker)
                "output_hidden_states": thinker_output_hidden_states,
                "return_dict_in_generate": True,
                # "output_scores": True,
                "eos_token_id": thinker_eos_token_ids,
                "pad_token_id": kwargs.get("thinker_pad_token_id", 151643),
            }
            model_inputs = {**inputs, **model_inputs}
            if len(thinker_stop_strings_per_step) > 0:
                model_inputs["stop_strings"] = thinker_stop_strings_per_step
                model_inputs["tokenizer"] = tokenizer

            start_time = perf_counter()
            outputs = self.thinker.generate(**model_inputs)
            times.append(perf_counter() - start_time)

            # Extract newly generated token IDs *for this step*
            # `outputs.sequences` contains the input_ids for this step + new tokens generated in this step
            step_new_ids = outputs.sequences[:, current_input_ids.shape[1] :]
            num_step_new_tokens = step_new_ids.shape[1]

            if num_step_new_tokens == 0:  # Handle case where generate stops early
                logging.warning("Warning: generate produced 0 new tokens in this step.")
                break

            if thinker_output_hidden_states is True:
                hidden_states = outputs.hidden_states
                hidden_states_len = (
                    hidden_states_len if hidden_states_len > 0 else hidden_states[0][0].shape[1]
                )
                logging.debug(f"hidden_states_len: {hidden_states_len}")
                # new generate thinker_token_embeds
                thinker_new_token_embeds = hidden_states[0][0][:, :hidden_states_len, :]
                hidden_states = (
                    (thinker_new_token_embeds,) + hidden_states[0][1:],
                ) + hidden_states[1:]
                # new generate thinker_hidden_states
                thinker_new_hidden_states = hidden_states[0][-1][:, :hidden_states_len, :]
                hidden_states = (
                    hidden_states[0][:-1] + (thinker_new_hidden_states,),
                ) + hidden_states[1:]

            yield {
                "thinker_generate_ids": step_new_ids,
                "thinker_generate_hidden_states": hidden_states,
            }
            total_new_tokens_generated += num_step_new_tokens

            # Update the full sequence
            full_generated_ids = torch.cat([full_generated_ids, step_new_ids], dim=1)

            # Prepare for the next iteration:
            # Input is only the last generated token
            # NOTE: need use past_key_values to keep the context by manually,
            # current_input_ids = step_new_ids[:, -1:]
            # so we can't use the last generated token, use cache instead
            # input ids need to be the full sequence for next generation
            current_input_ids = full_generated_ids

            # Update past_key_values
            # past_key_values = outputs.past_key_values

            # Update attention mask by appending 1s for the new tokens
            full_attention_mask = torch.cat(
                [full_attention_mask, torch.ones_like(step_new_ids)], dim=1
            )
            current_attention_mask = full_attention_mask

            # Check if EOS token was generated in this step
            if step_new_ids[0, -1].item() in thinker_eos_token_ids:
                logging.info("EOS token generated.")
                break

            # Check if max_new_tokens limit is reached (after processing the step)
            if total_new_tokens_generated >= thinker_max_new_tokens:
                logging.info("Max new tokens limit reached.")
                break

        logging.info(
            f"Total new tokens generated: {total_new_tokens_generated} | thinker_max_tokens_per_step: {thinker_max_tokens_per_step} | first chunk generated cost: {times[0]} s | total cost: {sum(times)} s"
        )

    @torch.no_grad()
    def talker_generate_chunk(
        self,
        inputs: dict,
        thinker_chunk_stream,
        speaker: str = "Chelsie",
        talker_eos_token_ids: list[int] = [8292, 8294],
        talker_top_k: int = 10,
        talker_top_p: float = 0.9,
        talker_temperature: float = 0.95,
        talker_repetition_penalty: float = 1.1,
        talker_min_new_tokens: int = 0,
        talker_max_new_tokens: int = 8192,
        talker_skip_thinker_token_ids: list[int] = [],  # skip tokens don't to talk
        code2wav_num_steps: int = 10,
        code2wav_guidance_scale: float = 0.5,
        code2wav_sway_coefficient: float = -1.0,
        code2wav_chunk_stream_func: Callable = None,
        mask_embedding: bool = True,
    ) -> Generator[dict, None, None]:
        input_ids = inputs.get("input_ids")
        attention_mask = inputs.get("attention_mask", None)

        for chunk in thinker_chunk_stream:
            thinker_generate_ids = chunk["thinker_generate_ids"].to(self.talker.device)
            # skip talk
            if (
                thinker_generate_ids.shape[1] == 1
                and thinker_generate_ids[0, -1].item() in talker_skip_thinker_token_ids
            ):
                logging.info(f"skip token {thinker_generate_ids} to talk")
                yield {"thinker_ids": thinker_generate_ids, "talker_wav": torch.empty([0])}
                continue
            thinker_generate_hidden_states = chunk["thinker_generate_hidden_states"]
            if thinker_generate_hidden_states is None or len(thinker_generate_hidden_states) < 2:
                if len(thinker_generate_hidden_states) < 2:
                    logging.warning(
                        f"thinker_generate_ids: {thinker_generate_ids} | len(thinker_generate_hidden_states): {len(thinker_generate_hidden_states)} < 2"
                    )
                yield {"thinker_ids": thinker_generate_ids, "talker_wav": torch.empty([0])}
                continue

            processed_thinker_hidden = thinker_generate_hidden_states
            if mask_embedding is True:
                logging.info("mask embedding")
                embeds_to_talker = (
                    thinker_generate_hidden_states[0][0].clone().to(self.talker.device)
                )
                if inputs.get("input_features", None) is not None:
                    audio_ids_mask = input_ids == self.config.thinker_config.audio_token_index
                    audio_mask = (
                        audio_ids_mask.unsqueeze(-1)
                        .expand_as(embeds_to_talker)
                        .to(embeds_to_talker.device)
                    )
                    audio_mask_tensor = torch.zeros(
                        [audio_ids_mask.sum(), embeds_to_talker.shape[-1]],
                        dtype=embeds_to_talker.dtype,
                        device=self.talker.device,
                    )
                    embeds_to_talker.masked_scatter_(audio_mask, audio_mask_tensor)
                if inputs.get("pixel_values", None) is not None:
                    image_ids_mask = input_ids == self.config.thinker_config.image_token_index
                    image_mask = (
                        image_ids_mask.unsqueeze(-1)
                        .expand_as(embeds_to_talker)
                        .to(embeds_to_talker.device)
                    )
                    image_mask_tensor = torch.zeros(
                        [image_ids_mask.sum(), embeds_to_talker.shape[-1]],
                        dtype=embeds_to_talker.dtype,
                        device=self.talker.device,
                    )
                    embeds_to_talker.masked_scatter_(image_mask, image_mask_tensor)
                if inputs.get("pixel_values_videos", None) is not None:
                    video_ids_mask = input_ids == self.config.thinker_config.video_token_index
                    video_mask = (
                        video_ids_mask.unsqueeze(-1)
                        .expand_as(embeds_to_talker)
                        .to(embeds_to_talker.device)
                    )
                    video_mask_tensor = torch.zeros(
                        [video_ids_mask.sum(), embeds_to_talker.shape[-1]],
                        dtype=embeds_to_talker.dtype,
                        device=self.talker.device,
                    )
                    embeds_to_talker.masked_scatter_(video_mask, video_mask_tensor)

                processed_thinker_hidden = (
                    (embeds_to_talker,) + thinker_generate_hidden_states[0][1:],
                ) + thinker_generate_hidden_states[1:]

            thinker_token_embeds = [
                token_hidden_states[0].to(self.talker.device)
                for token_hidden_states in processed_thinker_hidden
            ]
            thinker_hidden_states = [
                token_hidden_states[-1].to(self.talker.device)
                for token_hidden_states in processed_thinker_hidden
            ]
            logging.debug(
                f"len(thinker_generate_hidden_states):{len(thinker_generate_hidden_states)}"
            )
            for i in range(len(thinker_generate_hidden_states)):
                logging.debug(
                    f"thinker_generate_hidden_states[{i}]:{thinker_generate_hidden_states[i][0].shape}, {thinker_generate_hidden_states[i][-1].shape}"
                )

            talker_text_bos_token = self.speaker_map[speaker]["bos_token"]
            talker_input_text_ids = torch.cat(
                [
                    input_ids.to(self.talker.device),
                    torch.tensor(
                        [[talker_text_bos_token]], dtype=torch.long, device=self.talker.device
                    ),
                    thinker_generate_ids[:, :1],
                ],
                dim=-1,
            )
            logging.debug(f"talker_input_text_ids.shape:{talker_input_text_ids.shape}")

            talker_input_ids = torch.cat(
                [
                    torch.full_like(
                        input_ids,
                        fill_value=self.talker.codec_mask_token,
                        device=self.talker.device,
                    ),
                    torch.tensor(
                        [[self.talker.codec_pad_token]],
                        dtype=torch.long,
                        device=self.talker.device,
                    ),
                    torch.tensor(
                        [[self.talker.codec_bos_token]],
                        dtype=torch.long,
                        device=self.talker.device,
                    ),
                ],
                dim=1,
            )
            logging.debug(f"talker_input_ids.shape:{talker_input_ids.shape}")

            thinker_embed_tokens = self.thinker.get_input_embeddings()
            thinker_reply_part = torch.cat(thinker_hidden_states[1:], dim=1) + torch.cat(
                thinker_token_embeds[1:], dim=1
            )
            talker_inputs_embeds = thinker_hidden_states[0] + thinker_token_embeds[0]
            talker_text_bos_token = torch.tensor(
                [[talker_text_bos_token]], dtype=torch.long, device=self.thinker.device
            )
            talker_text_bos_embed = thinker_embed_tokens(talker_text_bos_token).to(
                self.talker.device
            )
            logging.debug(
                f"talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape} thinker_reply_part.shape {thinker_reply_part.shape}"
            )
            talker_inputs_embeds = torch.cat(
                [
                    talker_inputs_embeds,
                    talker_text_bos_embed,
                    thinker_reply_part[:, :1, :],
                ],
                dim=1,
            )
            logging.debug(
                f"talker_inputs_embeds.shape {talker_inputs_embeds.shape} talker_text_bos_embed.shape {talker_text_bos_embed.shape}"
            )

            eos_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[self.talker.text_eos_token]], dtype=torch.long, device=self.thinker.device
                )
            ).to(self.talker.device)

            pad_embedding = thinker_embed_tokens(
                torch.tensor(
                    [[self.talker.text_pad_token]], dtype=torch.long, device=self.thinker.device
                )
            ).to(self.talker.device)
            thinker_reply_part = torch.cat(
                [
                    thinker_reply_part[:, 1:, :],
                    eos_embedding,
                    pad_embedding,
                ],
                dim=1,
            )
            logging.debug(f"thinker_reply_part.shape:{thinker_reply_part.shape}")

            talker_attention_mask = None
            if attention_mask is not None:
                talker_attention_mask = torch.cat(
                    [attention_mask, attention_mask.new_ones((1, 2))], dim=1
                ).to(self.talker.device)

            streamer = TokenStreamer(skip_prompt=True)
            talker_kwargs = dict(
                input_ids=talker_input_ids,
                streamer=streamer,
                input_text_ids=talker_input_text_ids,
                thinker_reply_part=thinker_reply_part,
                inputs_embeds=talker_inputs_embeds,
                attention_mask=talker_attention_mask,
                suppress_tokens=[self.talker.codec_bos_token],
                eos_token_id=talker_eos_token_ids,
                pad_token_id=8292,
                do_sample=True if talker_temperature > 0.0 else False,
                top_k=talker_top_k,
                top_p=talker_top_p,
                temperature=talker_temperature,
                repetition_penalty=talker_repetition_penalty,
                min_new_tokens=talker_min_new_tokens,
                max_new_tokens=talker_max_new_tokens,
            )
            # logging.debug(talker_kwargs.keys())
            thread = Thread(target=self.talker.generate, kwargs=talker_kwargs)
            thread.start()

            code2wav_chunk_stream_func = code2wav_chunk_stream_func or self.code2wav_chunk_stream
            # Generate wavs from code
            code2wav_stream = code2wav_chunk_stream_func(
                talker_streamer=streamer,
                speaker=speaker,
                talker_eos_token_ids=talker_eos_token_ids,
                code2wav_num_steps=code2wav_num_steps,
                code2wav_guidance_scale=code2wav_guidance_scale,
                code2wav_sway_coefficient=code2wav_sway_coefficient,
            )

            for wav in code2wav_stream:
                yield {"thinker_ids": thinker_generate_ids, "talker_wav": wav}

    @torch.no_grad()
    def code2wav_chunk_stream(
        self,
        talker_streamer: TokenStreamer,
        speaker: str = "Chelsie",
        talker_eos_token_ids: list[int] = [8292, 8294],
        code2wav_num_steps: int = 10,
        code2wav_guidance_scale: float = 0.5,
        code2wav_sway_coefficient: float = -1.0,
    ) -> Generator[torch.Tensor, None, None]:
        """
        fixed chunk stream
        """
        if self.token2wav.dtype != torch.float:
            self.token2wav.float()

        code2wav_times = []
        talker_generate_codes = []
        times = []
        start_time = perf_counter()
        pre_offset = 0
        for token_id in talker_streamer:
            times.append(perf_counter() - start_time)
            start_time = perf_counter()
            if token_id in talker_eos_token_ids:
                break
            talker_generate_codes.append(token_id)
            chunk_code_length = len(talker_generate_codes) * 2 - 24
            if chunk_code_length > 0 and chunk_code_length % 48 == 0:
                codes_tensor = torch.tensor(
                    [talker_generate_codes[pre_offset:]],
                    dtype=torch.long,
                    device=self.talker.device,
                )
                pre_offset = len(talker_generate_codes)
                wav = self.token2wav(
                    codes_tensor.to(self.token2wav.device),
                    conditioning=self.speaker_map[speaker]["cond"]
                    .to(self.token2wav.device)
                    .float(),
                    reference_mel=self.speaker_map[speaker]["ref_mel"]
                    .to(self.token2wav.device)
                    .float(),
                    num_steps=10,
                    guidance_scale=0.5,
                    sway_coefficient=-1.0,
                ).detach()
                code2wav_times.append(perf_counter() - start_time)
                yield wav  # (T,)
                start_time = perf_counter()

        logging.info(
            f"talker generate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s"
        )

        if len(talker_generate_codes) > pre_offset:
            codes_tensor = torch.tensor(
                [talker_generate_codes[pre_offset:]],
                dtype=torch.long,
                device=self.talker.device,
            )
            wav = self.token2wav(
                codes_tensor.to(self.token2wav.device),
                conditioning=self.speaker_map[speaker]["cond"].to(self.token2wav.device).float(),
                reference_mel=self.speaker_map[speaker]["ref_mel"]
                .to(self.token2wav.device)
                .float(),
                num_steps=code2wav_num_steps,
                guidance_scale=code2wav_guidance_scale,
                sway_coefficient=code2wav_sway_coefficient,
            ).detach()
            code2wav_times.append(perf_counter() - start_time)
            yield wav  # (T,)

        logging.info(
            f"code2wav streaming first chunk time: {code2wav_times[0]} s | cost: {sum(code2wav_times)} s"
        )

    @torch.no_grad()
    def generate_stream(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
        thinker_max_new_tokens: int = 1024,
        thinker_top_k: int = 40,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_repetition_penalty: float = 1.05,
        thinker_eos_token_ids=[151643, 151645],
        thinker_stop_strings_per_step=[],
        tokenizer=None,
        return_audio=True,
        speaker="Chelsie",
        talker_top_k: int = 10,
        talker_top_p: float = 0.9,
        talker_temperature: float = 0.95,
        talker_repetition_penalty: float = 1.1,
        talker_min_new_tokens: int = 0,
        talker_max_new_tokens: int = 8192,
        talker_eos_token_ids: list[int] = [8292, 8294],
        talker_skip_thinker_token_ids: list[int] = [],
        code2wav_num_steps: int = 10,
        code2wav_guidance_scale: float = 0.5,
        code2wav_sway_coefficient: float = -1.0,
        code2wav_chunk_stream_func: Callable = None,
        mask_embedding: bool = True,
        **kwargs,
    ) -> Generator[dict, None, None]:
        """
        - return Generator[dict, None, None]
        {
            "thinker_ids": torch.Tensor, # (1,T)
            "talker_wav": torch.Tensor, # (T,)
        }
        """
        thinker_chunk_stream = self.thinker_generate_chunk(
            inputs,
            use_audio_in_video=use_audio_in_video,
            thinker_max_tokens_per_step=thinker_max_tokens_per_step,
            thinker_max_new_tokens=thinker_max_new_tokens,
            thinker_top_k=thinker_top_k,
            thinker_top_p=thinker_top_p,
            thinker_temperature=thinker_temperature,
            thinker_eos_token_ids=thinker_eos_token_ids,
            thinker_repetition_penalty=thinker_repetition_penalty,
            thinker_output_hidden_states=return_audio,
            thinker_stop_strings_per_step=thinker_stop_strings_per_step,
            tokenizer=tokenizer,
            **kwargs,
        )
        if not return_audio:
            for thinker_chunk in thinker_chunk_stream:
                yield {"thinker_ids": thinker_chunk["thinker_generate_ids"]}
        else:
            talker_streamer = self.talker_generate_chunk(
                inputs,
                thinker_chunk_stream=thinker_chunk_stream,
                speaker=speaker,
                talker_eos_token_ids=talker_eos_token_ids,
                talker_top_k=talker_top_k,
                talker_top_p=talker_top_p,
                talker_temperature=talker_temperature,
                talker_repetition_penalty=talker_repetition_penalty,
                talker_min_new_tokens=talker_min_new_tokens,
                talker_max_new_tokens=talker_max_new_tokens,
                talker_skip_thinker_token_ids=talker_skip_thinker_token_ids,
                code2wav_num_steps=code2wav_num_steps,
                code2wav_guidance_scale=code2wav_guidance_scale,
                code2wav_sway_coefficient=code2wav_sway_coefficient,
                code2wav_chunk_stream_func=code2wav_chunk_stream_func,
                mask_embedding=mask_embedding,
            )

            for talker_chunk in talker_streamer:
                yield talker_chunk

    @torch.no_grad()
    def thinker_all_talker_stream(
        self,
        inputs: dict,
        use_audio_in_video: bool = False,
        thinker_max_new_tokens: int = 1024,
        thinker_top_k: int = 40,
        thinker_top_p: float = 0.8,
        thinker_temperature: float = 0.9,
        thinker_repetition_penalty: float = 1.05,
        thinker_eos_token_ids=[151643, 151645],
        thinker_stop_strings_per_step=[],
        tokenizer=None,
        return_audio=True,
        speaker="Chelsie",
        talker_top_k: int = 10,
        talker_top_p: float = 0.9,
        talker_temperature: float = 0.95,
        talker_repetition_penalty: float = 1.1,
        talker_min_new_tokens: int = 0,
        talker_max_new_tokens: int = 8192,
        talker_eos_token_ids: list[int] = [8292, 8294],
        talker_skip_thinker_token_ids: list[int] = [],
        code2wav_num_steps: int = 10,
        code2wav_guidance_scale: float = 0.5,
        code2wav_sway_coefficient: float = -1.0,
        code2wav_chunk_stream_func: Callable = None,
        mask_embedding: bool = True,
        **kwargs,
    ) -> Generator[dict, None, None]:
        """
        - return Generator[dict, None, None]
        {
            "thinker_ids": torch.Tensor, # (1,T)
            "talker_wav": torch.Tensor, # (T,)
        }
        """

        def to_generator(item):
            yield item

        thinker_result = self.thinker.generate(
            **inputs,
            use_audio_in_video=use_audio_in_video,
            do_sample=True if thinker_temperature > 0 else False,
            top_k=thinker_top_k,
            top_p=thinker_top_p,
            temperature=thinker_temperature,
            repetition_penalty=thinker_repetition_penalty,
            min_new_tokens=1,
            max_new_tokens=thinker_max_new_tokens,
            eos_token_id=thinker_eos_token_ids,
            output_hidden_states=return_audio,
            return_dict_in_generate=True,
        )
        input_ids = inputs["input_ids"]
        thinker_generate_ids = thinker_result.sequences[:, input_ids.size(1) :]
        if not return_audio:
            yield {"thinker_ids": thinker_generate_ids}
        else:
            talker_streamer = self.talker_generate_chunk(
                inputs,
                thinker_chunk_stream=to_generator(
                    {
                        "thinker_generate_ids": thinker_generate_ids,
                        "thinker_generate_hidden_states": thinker_result.hidden_states,
                    }
                ),
                speaker=speaker,
                talker_eos_token_ids=talker_eos_token_ids,
                talker_top_k=talker_top_k,
                talker_top_p=talker_top_p,
                talker_temperature=talker_temperature,
                talker_repetition_penalty=talker_repetition_penalty,
                talker_min_new_tokens=talker_min_new_tokens,
                talker_max_new_tokens=talker_max_new_tokens,
                talker_skip_thinker_token_ids=talker_skip_thinker_token_ids,
                code2wav_num_steps=code2wav_num_steps,
                code2wav_guidance_scale=code2wav_guidance_scale,
                code2wav_sway_coefficient=code2wav_sway_coefficient,
                code2wav_chunk_stream_func=code2wav_chunk_stream_func,
                mask_embedding=mask_embedding,
            )

            for talker_chunk in talker_streamer:
                yield talker_chunk



from concurrent import futures
import grpc
import json
import time

import paraformer_pb2_grpc
from paraformer_pb2 import Response


class ASRServicer(paraformer_pb2_grpc.ASRServicer):
    def __init__(self, user_allowed, model, sample_rate, backend, onnx_dir, vad_model='', punc_model=''):
        print("ASRServicer init")
        self.backend = backend
        self.init_flag = 0
        self.client_buffers = {}
        self.client_transcription = {}
        self.auth_user = user_allowed.split("|")
        if self.backend == "pipeline":
            try:
                from modelscope.pipelines import pipeline
                from modelscope.utils.constant import Tasks
            except ImportError:
                raise ImportError(f"Please install modelscope")
            self.inference_16k_pipeline = pipeline(task=Tasks.auto_speech_recognition, model=model, vad_model=vad_model, punc_model=punc_model)
        elif self.backend == "onnxruntime":
            try:
                from funasr_local_onnx import Paraformer
            except ImportError:
                raise ImportError(f"Please install onnxruntime environment")
            self.inference_16k_pipeline = Paraformer(model_dir=onnx_dir)
        self.sample_rate = sample_rate

    def clear_states(self, user):
        self.clear_buffers(user)
        self.clear_transcriptions(user)

    def clear_buffers(self, user):
        if user in self.client_buffers:
            del self.client_buffers[user]

    def clear_transcriptions(self, user):
        if user in self.client_transcription:
            del self.client_transcription[user]

    def disconnect(self, user):
        self.clear_states(user)
        print("Disconnecting user: %s" % str(user))

    def Recognize(self, request_iterator, context):
        
            
        for req in request_iterator:
            if req.user not in self.auth_user:
                result = {}
                result["success"] = False
                result["detail"] = "Not Authorized user: %s " % req.user
                result["text"] = ""
                yield Response(sentence=json.dumps(result), user=req.user, action="terminate", language=req.language)
            elif req.isEnd: #end grpc
                print("asr end")
                self.disconnect(req.user)
                result = {}
                result["success"] = True
                result["detail"] = "asr end"
                result["text"] = ""
                yield Response(sentence=json.dumps(result), user=req.user, action="terminate",language=req.language)
            elif req.speaking: #continue speaking
                if req.audio_data is not None and len(req.audio_data) > 0:
                    if req.user in self.client_buffers:
                        self.client_buffers[req.user] += req.audio_data #append audio
                    else:
                        self.client_buffers[req.user] = req.audio_data
                result = {}
                result["success"] = True
                result["detail"] = "speaking"
                result["text"] = ""
                yield Response(sentence=json.dumps(result), user=req.user, action="speaking", language=req.language)
            elif not req.speaking: #silence
                if req.user not in self.client_buffers:
                    result = {}
                    result["success"] = True
                    result["detail"] = "waiting_for_more_voice"
                    result["text"] = ""
                    yield Response(sentence=json.dumps(result), user=req.user, action="waiting", language=req.language)
                else:
                    begin_time = int(round(time.time() * 1000))
                    tmp_data = self.client_buffers[req.user]
                    self.clear_states(req.user)
                    result = {}
                    result["success"] = True
                    result["detail"] = "decoding data: %d bytes" % len(tmp_data)
                    result["text"] = ""
                    yield Response(sentence=json.dumps(result), user=req.user, action="decoding", language=req.language)
                    if len(tmp_data) < 9600: #min input_len for asr model , 300ms
                        end_time = int(round(time.time() * 1000))
                        delay_str = str(end_time - begin_time)
                        result = {}
                        result["success"] = True
                        result["detail"] = "waiting_for_more_voice"
                        result["server_delay_ms"] = delay_str
                        result["text"] = ""
                        print ("user: %s , delay(ms): %s, info: %s " % (req.user, delay_str, "waiting_for_more_voice"))
                        yield Response(sentence=json.dumps(result), user=req.user, action="waiting", language=req.language)
                    else:
                        if self.backend == "pipeline":
                            asr_result = self.inference_16k_pipeline(audio_in=tmp_data, audio_fs = self.sample_rate)
                            if "text" in asr_result:
                                asr_result = asr_result['text']
                            else:
                                asr_result = ""
                        elif self.backend == "onnxruntime":
                            from funasr_local_onnx.utils.frontend import load_bytes
                            array = load_bytes(tmp_data)
                            asr_result = self.inference_16k_pipeline(array)[0]
                        end_time = int(round(time.time() * 1000))
                        delay_str = str(end_time - begin_time)
                        print ("user: %s , delay(ms): %s, text: %s " % (req.user, delay_str, asr_result))
                        result = {}
                        result["success"] = True
                        result["detail"] = "finish_sentence"
                        result["server_delay_ms"] = delay_str
                        result["text"] = asr_result
                        yield Response(sentence=json.dumps(result), user=req.user, action="finish", language=req.language)
            else:
                result = {}
                result["success"] = False 
                result["detail"] = "error, no condition matched! Unknown reason."
                result["text"] = ""
                self.disconnect(req.user)
                yield Response(sentence=json.dumps(result), user=req.user, action="terminate", language=req.language)
                




# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional

import torch
import torch.nn as nn
from fairseq import utils
from fairseq.modules import LayerNorm, MultiheadAttention
from fairseq.modules.fairseq_dropout import FairseqDropout
from fairseq.modules.quant_noise import quant_noise


class TransformerSentenceEncoderLayer(nn.Module):
    """
    Implements a Transformer Encoder Layer used in BERT/XLM style pre-trained
    models.
    """

    def __init__(
        self,
        embedding_dim: int = 768,
        ffn_embedding_dim: int = 3072,
        num_attention_heads: int = 8,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation_dropout: float = 0.1,
        activation_fn: str = "relu",
        export: bool = False,
        q_noise: float = 0.0,
        qn_block_size: int = 8,
        init_fn: Callable = None,
    ) -> None:
        super().__init__()

        if init_fn is not None:
            init_fn()

        # Initialize parameters
        self.embedding_dim = embedding_dim
        self.num_attention_heads = num_attention_heads
        self.attention_dropout = attention_dropout
        self.q_noise = q_noise
        self.qn_block_size = qn_block_size

        self.dropout_module = FairseqDropout(
            dropout, module_name=self.__class__.__name__
        )
        self.activation_dropout_module = FairseqDropout(
            activation_dropout, module_name=self.__class__.__name__
        )

        # Initialize blocks
        self.activation_fn = utils.get_activation_fn(activation_fn)
        self.self_attn = self.build_self_attention(
            self.embedding_dim,
            num_attention_heads,
            dropout=attention_dropout,
            self_attention=True,
            q_noise=q_noise,
            qn_block_size=qn_block_size,
        )

        # layer norm associated with the self attention layer
        self.self_attn_layer_norm = LayerNorm(self.embedding_dim, export=export)

        self.fc1 = self.build_fc1(
            self.embedding_dim,
            ffn_embedding_dim,
            q_noise=q_noise,
            qn_block_size=qn_block_size,
        )
        self.fc2 = self.build_fc2(
            ffn_embedding_dim,
            self.embedding_dim,
            q_noise=q_noise,
            qn_block_size=qn_block_size,
        )

        # layer norm associated with the position wise feed-forward NN
        self.final_layer_norm = LayerNorm(self.embedding_dim, export=export)

    def build_fc1(self, input_dim, output_dim, q_noise, qn_block_size):
        return quant_noise(nn.Linear(input_dim, output_dim), q_noise, qn_block_size)

    def build_fc2(self, input_dim, output_dim, q_noise, qn_block_size):
        return quant_noise(nn.Linear(input_dim, output_dim), q_noise, qn_block_size)

    def build_self_attention(
        self,
        embed_dim,
        num_attention_heads,
        dropout,
        self_attention,
        q_noise,
        qn_block_size,
    ):
        return MultiheadAttention(
            embed_dim,
            num_attention_heads,
            dropout=dropout,
            self_attention=True,
            q_noise=q_noise,
            qn_block_size=qn_block_size,
        )

    def forward(
        self,
        x: torch.Tensor,
        self_attn_mask: Optional[torch.Tensor] = None,
        self_attn_padding_mask: Optional[torch.Tensor] = None,
    ):
        """
        LayerNorm is applied either before or after the self-attention/ffn
        modules similar to the original Transformer implementation.
        """
        residual = x
        x, attn = self.self_attn(
            query=x,
            key=x,
            value=x,
            key_padding_mask=self_attn_padding_mask,
            need_weights=False,
            attn_mask=self_attn_mask,
        )
        x = self.dropout_module(x)
        x = residual + x
        x = self.self_attn_layer_norm(x)

        residual = x
        x = self.activation_fn(self.fc1(x))
        x = self.activation_dropout_module(x)
        x = self.fc2(x)
        x = self.dropout_module(x)
        x = residual + x
        x = self.final_layer_norm(x)
        return x, attn



import os
import numpy as np
import random
from PIL import Image
import torch
from torch.utils.data import Dataset, ConcatDataset
import torchvision.transforms as transforms
from transformers import AutoFeatureExtractor
import librosa
import time
import json
import math
from decord import AudioReader, VideoReader
from decord.ndarray import cpu

from musetalk.data.sample_method import get_src_idx, shift_landmarks_to_face_coordinates, resize_landmark 
from musetalk.data import audio 

syncnet_mel_step_size = math.ceil(16 / 5 * 16)  # latentsync


class FaceDataset(Dataset):
    """Dataset class for loading and processing video data
    
    Each video can be represented as:
    - Concatenated frame images
    - '.mp4' or '.gif' files
    - Folder containing all frames
    """
    def __init__(self,
                 cfg,
                 list_paths,
                 root_path='./dataset/',
                 repeats=None):
        # Initialize dataset paths
        meta_paths = []
        if repeats is None:
            repeats = [1] * len(list_paths)
        assert len(repeats) == len(list_paths)
        
        # Load data list
        for list_path, repeat_time in zip(list_paths, repeats):
            with open(list_path, 'r') as f:
                num = 0
                f.readline()  # Skip header line
                for line in f.readlines():
                    line_info = line.strip()
                    meta = line_info.split()
                    meta = meta[0]
                    meta_paths.extend([os.path.join(root_path, meta)] * repeat_time)
                    num += 1
                print(f'{list_path}: {num} x {repeat_time} = {num * repeat_time} samples')

        # Set basic attributes
        self.meta_paths = meta_paths
        self.root_path = root_path
        self.image_size = cfg['image_size']
        self.min_face_size = cfg['min_face_size']
        self.T = cfg['T']
        self.sample_method = cfg['sample_method']
        self.top_k_ratio = cfg['top_k_ratio']
        self.max_attempts = 200
        self.padding_pixel_mouth = cfg['padding_pixel_mouth']
        
        # Cropping related parameters
        self.crop_type = cfg['crop_type']
        self.jaw2edge_margin_mean = cfg['cropping_jaw2edge_margin_mean']
        self.jaw2edge_margin_std = cfg['cropping_jaw2edge_margin_std']
        self.random_margin_method = cfg['random_margin_method']
        
        # Image transformations
        self.to_tensor = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
        ])
        self.pose_to_tensor = transforms.Compose([
            transforms.ToTensor(),
        ])

        # Feature extractor
        self.feature_extractor = AutoFeatureExtractor.from_pretrained(cfg['whisper_path'])
        self.contorl_face_min_size = cfg["contorl_face_min_size"]
        
        print("The sample method is: ", self.sample_method)
        print(f"only use face size > {self.min_face_size}", self.contorl_face_min_size)

    def generate_random_value(self):
        """Generate random value
        
        Returns:
            float: Generated random value
        """
        if self.random_margin_method == "uniform":
            random_value = np.random.uniform(
                self.jaw2edge_margin_mean - self.jaw2edge_margin_std, 
                self.jaw2edge_margin_mean + self.jaw2edge_margin_std
            )
        elif self.random_margin_method == "normal":
            random_value = np.random.normal(
                loc=self.jaw2edge_margin_mean, 
                scale=self.jaw2edge_margin_std
            )
            random_value = np.clip(
                random_value, 
                self.jaw2edge_margin_mean - self.jaw2edge_margin_std, 
                self.jaw2edge_margin_mean + self.jaw2edge_margin_std, 
            )
        else:
            raise ValueError(f"Invalid random margin method: {self.random_margin_method}")
        return max(0, random_value)

    def dynamic_margin_crop(self, img, original_bbox, extra_margin=None):
        """Dynamically crop image with dynamic margin
        
        Args:
            img: Input image
            original_bbox: Original bounding box
            extra_margin: Extra margin
            
        Returns:
            tuple: (x1, y1, x2, y2, extra_margin)
        """
        if extra_margin is None:
            extra_margin = self.generate_random_value()
        w, h = img.size
        x1, y1, x2, y2 = original_bbox
        y2 = min(y2 + int(extra_margin), h)
        return x1, y1, x2, y2, extra_margin

    def crop_resize_img(self, img, bbox, crop_type='crop_resize', extra_margin=None):
        """Crop and resize image
        
        Args:
            img: Input image
            bbox: Bounding box
            crop_type: Type of cropping
            extra_margin: Extra margin
            
        Returns:
            tuple: (Processed image, extra_margin, mask_scaled_factor)
        """
        mask_scaled_factor = 1.
        if crop_type == 'crop_resize':
            x1, y1, x2, y2 = bbox
            img = img.crop((x1, y1, x2, y2))
            img = img.resize((self.image_size, self.image_size), Image.LANCZOS)
        elif crop_type == 'dynamic_margin_crop_resize':
            x1, y1, x2, y2, extra_margin = self.dynamic_margin_crop(img, bbox, extra_margin)
            w_original, _ = img.size
            img = img.crop((x1, y1, x2, y2))
            w_cropped, _ = img.size
            mask_scaled_factor = w_cropped / w_original
            img = img.resize((self.image_size, self.image_size), Image.LANCZOS)
        elif crop_type == 'resize':
            w, h = img.size
            scale = np.sqrt(self.image_size ** 2 / (h * w))
            new_w = int(w * scale) / 64 * 64
            new_h = int(h * scale) / 64 * 64
            img = img.resize((new_w, new_h), Image.LANCZOS)
        return img, extra_margin, mask_scaled_factor

    def get_audio_file(self, wav_path, start_index):
        """Get audio file features
        
        Args:
            wav_path: Audio file path
            start_index: Starting index
            
        Returns:
            tuple: (Audio features, start index)
        """
        if not os.path.exists(wav_path):
            return None
        audio_input_librosa, sampling_rate = librosa.load(wav_path, sr=16000)
        assert sampling_rate == 16000

        while start_index >= 25 * 30:
            audio_input = audio_input_librosa[16000*30:]
            start_index -= 25 * 30
        if start_index + 2 * 25 >= 25 * 30:
            start_index -= 4 * 25
            audio_input = audio_input_librosa[16000*4:16000*34]
        else:
            audio_input = audio_input_librosa[:16000*30]

        assert 2 * (start_index) >= 0
        assert 2 * (start_index + 2 * 25) <= 1500

        audio_input = self.feature_extractor(
            audio_input,
            return_tensors="pt",
            sampling_rate=sampling_rate
        ).input_features
        return audio_input, start_index

    def get_audio_file_mel(self, wav_path, start_index):
        """Get mel spectrogram of audio file
        
        Args:
            wav_path: Audio file path
            start_index: Starting index
            
        Returns:
            tuple: (Mel spectrogram, start index)
        """
        if not os.path.exists(wav_path):
            return None

        audio_input, sampling_rate = librosa.load(wav_path, sr=16000)
        assert sampling_rate == 16000

        audio_input = self.mel_feature_extractor(audio_input)
        return audio_input, start_index

    def mel_feature_extractor(self, audio_input):
        """Extract mel spectrogram features
        
        Args:
            audio_input: Input audio
            
        Returns:
            ndarray: Mel spectrogram features
        """
        orig_mel = audio.melspectrogram(audio_input)
        return orig_mel.T

    def crop_audio_window(self, spec, start_frame_num, fps=25):
        """Crop audio window
        
        Args:
            spec: Spectrogram
            start_frame_num: Starting frame number
            fps: Frames per second
            
        Returns:
            ndarray: Cropped spectrogram
        """
        start_idx = int(80. * (start_frame_num / float(fps)))
        end_idx = start_idx + syncnet_mel_step_size
        return spec[start_idx: end_idx, :]

    def get_syncnet_input(self, video_path):
        """Get SyncNet input features
        
        Args:
            video_path: Video file path
            
        Returns:
            ndarray: SyncNet input features
        """
        ar = AudioReader(video_path, sample_rate=16000)
        original_mel = audio.melspectrogram(ar[:].asnumpy().squeeze(0))
        return original_mel.T

    def get_resized_mouth_mask(
        self, 
        img_resized, 
        landmark_array, 
        face_shape, 
        padding_pixel_mouth=0, 
        image_size=256,
        crop_margin=0
    ):
        landmark_array = np.array(landmark_array)
        resized_landmark = resize_landmark(
            landmark_array, w=face_shape[0], h=face_shape[1], new_w=image_size, new_h=image_size)

        landmark_array = np.array(resized_landmark[48 : 67])  # the lip landmarks in 68 landmarks format
        min_x, min_y = np.min(landmark_array, axis=0)
        max_x, max_y = np.max(landmark_array, axis=0)
        min_x = min_x - padding_pixel_mouth
        max_x = max_x + padding_pixel_mouth

        # Calculate x-axis length and use it for y-axis
        width = max_x - min_x

        # Calculate old center point
        center_y = (max_y + min_y) / 2

        # Determine new min_y and max_y based on width
        min_y = center_y - width / 4
        max_y = center_y + width / 4

        # Adjust mask position for dynamic crop, shift y-axis
        min_y = min_y - crop_margin
        max_y = max_y - crop_margin
        
        # Prevent out of bounds
        min_x = max(min_x, 0)
        min_y = max(min_y, 0)
        max_x = min(max_x, face_shape[0])
        max_y = min(max_y, face_shape[1])

        mask = np.zeros_like(np.array(img_resized))
        mask[round(min_y):round(max_y), round(min_x):round(max_x)] = 255
        return Image.fromarray(mask)

    def __len__(self):
        return 100000

    def __getitem__(self, idx):
        attempts = 0
        while attempts < self.max_attempts:
            try:
                meta_path = random.sample(self.meta_paths, k=1)[0]
                with open(meta_path, 'r') as f:
                    meta_data = json.load(f)
            except Exception as e:
                print(f"meta file error:{meta_path}")
                print(e)
                attempts += 1
                time.sleep(0.1)
                continue
            
            video_path = meta_data["mp4_path"]
            wav_path =  meta_data["wav_path"]
            bbox_list = meta_data["face_list"]
            landmark_list = meta_data["landmark_list"]
            T = self.T

            s = 0
            e = meta_data["frames"]
            len_valid_clip = e - s

            if len_valid_clip < T * 10:
                attempts += 1
                print(f"video {video_path} has less than {T * 10} frames")
                continue

            try:
                cap = VideoReader(video_path, fault_tol=1, ctx=cpu(0))
                total_frames = len(cap)
                assert total_frames == len(landmark_list)
                assert total_frames == len(bbox_list)
                landmark_shape = np.array(landmark_list).shape
                if landmark_shape != (total_frames, 68, 2):
                    attempts += 1
                    print(f"video {video_path} has invalid landmark shape: {landmark_shape}, expected: {(total_frames, 68, 2)}") # we use 68 landmarks     
                    continue
            except Exception as e:
                print(f"video file error:{video_path}")
                print(e)
                attempts += 1
                time.sleep(0.1)
                continue

            shift_landmarks, bbox_list_union, face_shapes = shift_landmarks_to_face_coordinates(
                landmark_list, 
                bbox_list
            )
            if self.contorl_face_min_size and face_shapes[0][0] < self.min_face_size:
                print(f"video {video_path} has face size {face_shapes[0][0]} less than minimum required {self.min_face_size}")
                attempts += 1
                continue
                
            step = 1
            drive_idx_start = random.randint(s, e - T * step)
            drive_idx_list = list(
                range(drive_idx_start, drive_idx_start + T * step, step))
            assert len(drive_idx_list) == T

            src_idx_list = []
            list_index_out_of_range = False
            for drive_idx in drive_idx_list:
                src_idx = get_src_idx(
                    drive_idx, T, self.sample_method, shift_landmarks, face_shapes, self.top_k_ratio)
                if src_idx is None:
                    list_index_out_of_range = True
                    break
                src_idx = min(src_idx, e - 1)
                src_idx = max(src_idx, s)
                src_idx_list.append(src_idx)

            if list_index_out_of_range:
                attempts += 1
                print(f"video {video_path} has invalid source index for drive frames")
                continue

            ref_face_valid_flag = True
            extra_margin = self.generate_random_value()
            
            # Get reference images
            ref_imgs = []
            for src_idx in src_idx_list:
                imSrc = Image.fromarray(cap[src_idx].asnumpy())
                bbox_s = bbox_list_union[src_idx]
                imSrc, _, _ = self.crop_resize_img(
                    imSrc,
                    bbox_s, 
                    self.crop_type, 
                    extra_margin=None
                )
                if self.contorl_face_min_size and min(imSrc.size[0], imSrc.size[1]) < self.min_face_size:
                    ref_face_valid_flag = False
                    break
                ref_imgs.append(imSrc)

            if not ref_face_valid_flag:
                attempts += 1
                print(f"video {video_path} has reference face size smaller than minimum required {self.min_face_size}")
                continue
            
            # Get target images and masks
            imSameIDs = []
            bboxes = []
            face_masks = []
            face_mask_valid = True
            target_face_valid_flag = True
            
            for drive_idx in drive_idx_list:
                imSameID = Image.fromarray(cap[drive_idx].asnumpy())
                bbox_s = bbox_list_union[drive_idx]
                imSameID, _ , mask_scaled_factor = self.crop_resize_img(
                    imSameID, 
                    bbox_s, 
                    self.crop_type, 
                    extra_margin=extra_margin
                )
                if self.contorl_face_min_size and min(imSameID.size[0], imSameID.size[1]) < self.min_face_size:
                    target_face_valid_flag = False
                    break
                crop_margin = extra_margin * mask_scaled_factor
                face_mask = self.get_resized_mouth_mask(
                    imSameID,
                    shift_landmarks[drive_idx],
                    face_shapes[drive_idx],
                    self.padding_pixel_mouth,
                    self.image_size,
                    crop_margin=crop_margin
                )
                if np.count_nonzero(face_mask) == 0:
                    face_mask_valid = False
                    break

                if face_mask.size[1] == 0 or face_mask.size[0] == 0:
                    print(f"video {video_path} has invalid face mask size at frame {drive_idx}")
                    face_mask_valid = False
                    break

                imSameIDs.append(imSameID)
                bboxes.append(bbox_s)
                face_masks.append(face_mask)

            if not face_mask_valid:
                attempts += 1
                print(f"video {video_path} has invalid face mask")
                continue

            if not target_face_valid_flag:
                attempts += 1
                print(f"video {video_path} has target face size smaller than minimum required {self.min_face_size}")
                continue

            # Process audio features
            audio_offset = drive_idx_list[0]
            audio_step = step
            fps = 25.0 / step

            try:
                audio_feature, audio_offset = self.get_audio_file(wav_path, audio_offset)
                _, audio_offset = self.get_audio_file_mel(wav_path, audio_offset)
                audio_feature_mel = self.get_syncnet_input(video_path)
            except Exception as e:
                print(f"audio file error:{wav_path}")
                print(e)
                attempts += 1
                time.sleep(0.1)
                continue
            
            mel = self.crop_audio_window(audio_feature_mel, audio_offset)
            if mel.shape[0] != syncnet_mel_step_size:
                attempts += 1
                print(f"video {video_path} has invalid mel spectrogram shape: {mel.shape}, expected: {syncnet_mel_step_size}")
                continue
                
            mel = torch.FloatTensor(mel.T).unsqueeze(0)
            
            # Build sample dictionary
            sample = dict(
                pixel_values_vid=torch.stack(
                    [self.to_tensor(imSameID) for imSameID in imSameIDs], dim=0),
                pixel_values_ref_img=torch.stack(
                    [self.to_tensor(ref_img) for ref_img in ref_imgs], dim=0),
                pixel_values_face_mask=torch.stack(
                    [self.pose_to_tensor(face_mask) for face_mask in face_masks], dim=0),
                audio_feature=audio_feature[0],
                audio_offset=audio_offset,
                audio_step=audio_step,
                mel=mel,
                wav_path=wav_path,
                fps=fps,
            )

            return sample

        raise ValueError("Unable to find a valid sample after maximum attempts.")

class HDTFDataset(FaceDataset):
    """HDTF dataset class"""
    def __init__(self, cfg):
        root_path = './dataset/HDTF/meta'
        list_paths = [
            './dataset/HDTF/train.txt',
        ]
        

        repeats = [10]
        super().__init__(cfg, list_paths, root_path, repeats)
        print('HDTFDataset: ', len(self))

class VFHQDataset(FaceDataset):
    """VFHQ dataset class"""
    def __init__(self, cfg):
        root_path = './dataset/VFHQ/meta'
        list_paths = [
            './dataset/VFHQ/train.txt',
        ]
        repeats = [1]
        super().__init__(cfg, list_paths, root_path, repeats)
        print('VFHQDataset: ', len(self))
        
def PortraitDataset(cfg=None):
    """Return dataset based on configuration
    
    Args:
        cfg: Configuration dictionary
        
    Returns:
        Dataset: Combined dataset
    """
    if cfg["dataset_key"] == "HDTF":
        return ConcatDataset([HDTFDataset(cfg)])
    elif cfg["dataset_key"] == "VFHQ":
        return ConcatDataset([VFHQDataset(cfg)])
    else:  
        print("############ use all dataset ############ ")
        return ConcatDataset([HDTFDataset(cfg), VFHQDataset(cfg)])


if __name__ == '__main__':
    # Set random seeds for reproducibility
    seed = 42
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    # Create dataset with configuration parameters
    dataset = PortraitDataset(cfg={
        'T': 1,  # Number of frames to process at once
        'random_margin_method': "normal",  # Method for generating random margins: "normal" or "uniform"
        'dataset_key': "HDTF",  # Dataset to use: "HDTF", "VFHQ", or None for both
        'image_size': 256,  # Size of processed images (height and width)
        'sample_method': 'pose_similarity_and_mouth_dissimilarity',  # Method for selecting reference frames
        'top_k_ratio': 0.51,  # Ratio for top-k selection in reference frame sampling
        'contorl_face_min_size': True,  # Whether to enforce minimum face size
        'padding_pixel_mouth': 10,  # Padding pixels around mouth region in mask
        'min_face_size': 200,  # Minimum face size requirement for dataset
        'whisper_path': "./models/whisper",  # Path to Whisper model
        'cropping_jaw2edge_margin_mean': 10,  # Mean margin for jaw-to-edge cropping
        'cropping_jaw2edge_margin_std': 10,  # Standard deviation for jaw-to-edge cropping
        'crop_type': "dynamic_margin_crop_resize",  # Type of cropping: "crop_resize", "dynamic_margin_crop_resize", or "resize"
    })
    print(len(dataset))
    
    import torchvision
    os.makedirs('debug', exist_ok=True)
    for i in range(10):  # Check 10 samples
        sample = dataset[0]
        print(f"processing {i}")
        
        # Get images and mask
        ref_img = (sample['pixel_values_ref_img'] + 1.0) / 2  # (b, c, h, w)
        target_img = (sample['pixel_values_vid'] + 1.0) / 2
        face_mask = sample['pixel_values_face_mask']
        
        # Print dimension information
        print(f"ref_img shape: {ref_img.shape}")
        print(f"target_img shape: {target_img.shape}")
        print(f"face_mask shape: {face_mask.shape}")
        
        # Create visualization images
        b, c, h, w = ref_img.shape
        
        # Apply mask only to target image
        target_mask = face_mask
        
        # Keep reference image unchanged
        ref_with_mask = ref_img.clone()
        
        # Create mask overlay for target image
        target_with_mask = target_img.clone()
        target_with_mask = target_with_mask * (1 - target_mask) + target_mask  # Apply mask only to target
        
        # Save original images, mask, and overlay results
        # First row: original images
        # Second row: mask
        # Third row: overlay effect
        concatenated_img = torch.cat((
            ref_img, target_img,  # Original images
            torch.zeros_like(ref_img), target_mask,  # Mask (black for ref)
            ref_with_mask, target_with_mask  # Overlay effect
        ), dim=3)
        
        torchvision.utils.save_image(
            concatenated_img, f'debug/mask_check_{i}.jpg', nrow=2)



import logging
from pathlib import Path


class Logger:
    def __init__(self, logger=logging.getLogger(Path(__file__).parent.name)):
        self.logger = logger

    def set_logger(self, logger: logging.Logger):
        self.logger = logger

    def get_logger(self) -> logging.Logger:
        return self.logger


logger = Logger()



import sys
from face_detection import FaceAlignment,LandmarksType
from os import listdir, path
import subprocess
import numpy as np
import cv2
import pickle
import os
import json
from mmpose.apis import inference_topdown, init_model
from mmpose.structures import merge_data_samples
import torch
from tqdm import tqdm

# initialize the mmpose model
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
config_file = './musetalk/utils/dwpose/rtmpose-l_8xb32-270e_coco-ubody-wholebody-384x288.py'
checkpoint_file = './models/dwpose/dw-ll_ucoco_384.pth'
model = init_model(config_file, checkpoint_file, device=device)

# initialize the face detection model
device = "cuda" if torch.cuda.is_available() else "cpu"
fa = FaceAlignment(LandmarksType._2D, flip_input=False,device=device)

# maker if the bbox is not sufficient 
coord_placeholder = (0.0,0.0,0.0,0.0)

def resize_landmark(landmark, w, h, new_w, new_h):
    w_ratio = new_w / w
    h_ratio = new_h / h
    landmark_norm = landmark / [w, h]
    landmark_resized = landmark_norm * [new_w, new_h]
    return landmark_resized

def read_imgs(img_list):
    frames = []
    print('reading images...')
    for img_path in tqdm(img_list):
        frame = cv2.imread(img_path)
        frames.append(frame)
    return frames

def get_bbox_range(img_list,upperbondrange =0):
    frames = read_imgs(img_list)
    batch_size_fa = 1
    batches = [frames[i:i + batch_size_fa] for i in range(0, len(frames), batch_size_fa)]
    coords_list = []
    landmarks = []
    if upperbondrange != 0:
        print('get key_landmark and face bounding boxes with the bbox_shift:',upperbondrange)
    else:
        print('get key_landmark and face bounding boxes with the default value')
    average_range_minus = []
    average_range_plus = []
    for fb in tqdm(batches):
        results = inference_topdown(model, np.asarray(fb)[0])
        results = merge_data_samples(results)
        keypoints = results.pred_instances.keypoints
        face_land_mark= keypoints[0][23:91]
        face_land_mark = face_land_mark.astype(np.int32)
        
        # get bounding boxes by face detetion
        bbox = fa.get_detections_for_batch(np.asarray(fb))
        
        # adjust the bounding box refer to landmark
        # Add the bounding box to a tuple and append it to the coordinates list
        for j, f in enumerate(bbox):
            if f is None: # no face in the image
                coords_list += [coord_placeholder]
                continue
            
            half_face_coord =  face_land_mark[29]#np.mean([face_land_mark[28], face_land_mark[29]], axis=0)
            range_minus = (face_land_mark[30]- face_land_mark[29])[1]
            range_plus = (face_land_mark[29]- face_land_mark[28])[1]
            average_range_minus.append(range_minus)
            average_range_plus.append(range_plus)
            if upperbondrange != 0:
                half_face_coord[1] = upperbondrange+half_face_coord[1] #手动调整  + 向下（偏29）  - 向上（偏28）

    text_range=f"Total frame:「{len(frames)}」 Manually adjust range : [ -{int(sum(average_range_minus) / len(average_range_minus))}~{int(sum(average_range_plus) / len(average_range_plus))} ] , the current value: {upperbondrange}"
    return text_range
    

def get_landmark_and_bbox(img_list,upperbondrange =0):
    frames = read_imgs(img_list)
    batch_size_fa = 1
    batches = [frames[i:i + batch_size_fa] for i in range(0, len(frames), batch_size_fa)]
    coords_list = []
    landmarks = []
    if upperbondrange != 0:
        print('get key_landmark and face bounding boxes with the bbox_shift:',upperbondrange)
    else:
        print('get key_landmark and face bounding boxes with the default value')
    average_range_minus = []
    average_range_plus = []
    for fb in tqdm(batches):
        results = inference_topdown(model, np.asarray(fb)[0])
        results = merge_data_samples(results)
        keypoints = results.pred_instances.keypoints
        face_land_mark= keypoints[0][23:91]
        face_land_mark = face_land_mark.astype(np.int32)
        
        # get bounding boxes by face detetion
        bbox = fa.get_detections_for_batch(np.asarray(fb))
        
        # adjust the bounding box refer to landmark
        # Add the bounding box to a tuple and append it to the coordinates list
        for j, f in enumerate(bbox):
            if f is None: # no face in the image
                coords_list += [coord_placeholder]
                continue
            
            half_face_coord =  face_land_mark[29]#np.mean([face_land_mark[28], face_land_mark[29]], axis=0)
            range_minus = (face_land_mark[30]- face_land_mark[29])[1]
            range_plus = (face_land_mark[29]- face_land_mark[28])[1]
            average_range_minus.append(range_minus)
            average_range_plus.append(range_plus)
            if upperbondrange != 0:
                half_face_coord[1] = upperbondrange+half_face_coord[1] #手动调整  + 向下（偏29）  - 向上（偏28）
            half_face_dist = np.max(face_land_mark[:,1]) - half_face_coord[1]
            upper_bond = half_face_coord[1]-half_face_dist
            
            f_landmark = (np.min(face_land_mark[:, 0]),int(upper_bond),np.max(face_land_mark[:, 0]),np.max(face_land_mark[:,1]))
            x1, y1, x2, y2 = f_landmark
            
            if y2-y1<=0 or x2-x1<=0 or x1<0: # if the landmark bbox is not suitable, reuse the bbox
                coords_list += [f]
                w,h = f[2]-f[0], f[3]-f[1]
                print("error bbox:",f)
            else:
                coords_list += [f_landmark]
    
    print("********************************************bbox_shift parameter adjustment**********************************************************")
    print(f"Total frame:「{len(frames)}」 Manually adjust range : [ -{int(sum(average_range_minus) / len(average_range_minus))}~{int(sum(average_range_plus) / len(average_range_plus))} ] , the current value: {upperbondrange}")
    print("*************************************************************************************************************************************")
    return coords_list,frames
    

if __name__ == "__main__":
    img_list = ["./results/lyria/00000.png","./results/lyria/00001.png","./results/lyria/00002.png","./results/lyria/00003.png"]
    crop_coord_path = "./coord_face.pkl"
    coords_list,full_frames = get_landmark_and_bbox(img_list)
    with open(crop_coord_path, 'wb') as f:
        pickle.dump(coords_list, f)
        
    for bbox, frame in zip(coords_list,full_frames):
        if bbox == coord_placeholder:
            continue
        x1, y1, x2, y2 = bbox
        crop_frame = frame[y1:y2, x1:x2]
        print('Cropped shape', crop_frame.shape)
        
        #cv2.imwrite(path.join(save_dir, '{}.png'.format(i)),full_frames[i][0][y1:y2, x1:x2])
    print(coords_list)



# Copyright 2025 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import torch
from torch import nn
import torch.nn.functional as F

from tts.modules.wavvae.decoder.seanet_encoder import Encoder
from tts.modules.wavvae.decoder.diag_gaussian import DiagonalGaussianDistribution
from tts.modules.wavvae.decoder.hifigan_modules import Generator, Upsample


class WavVAE_V3(nn.Module):
    def __init__(self, hparams=None):
        super().__init__()
        self.encoder = Encoder(dowmsamples=[6, 5, 4, 4, 2])
        self.proj_to_z = nn.Linear(512, 64)
        self.proj_to_decoder = nn.Linear(32, 320)

        config_path = hparams['melgan_config']
        args = argparse.Namespace()
        args.__dict__.update(config_path)
        self.latent_upsampler = Upsample(320, 4)
        self.decoder = Generator(
            input_size_=160, ngf=128, n_residual_layers=4,
            num_band=1, args=args, ratios=[5,4,4,3])

    ''' encode waveform into 25 hz latent representation '''
    def encode_latent(self, audio):
        posterior = self.encode(audio)
        latent = posterior.sample().permute(0, 2, 1)  # (b,t,latent_channel)
        return latent

    def encode(self, audio):
        x = self.encoder(audio).permute(0, 2, 1)
        x = self.proj_to_z(x).permute(0, 2, 1)
        poseterior = DiagonalGaussianDistribution(x)
        return poseterior

    def decode(self, latent):
        latent = self.proj_to_decoder(latent).permute(0, 2, 1)
        return self.decoder(self.latent_upsampler(latent))

    def forward(self, audio):
        posterior = self.encode(audio)
        latent = posterior.sample().permute(0, 2, 1)  # (b, t, latent_channel)
        recon_wav = self.decode(latent)
        return recon_wav, posterior


import asyncio
import logging
from string import Template

from transformers import AutoTokenizer
from apipeline.frames.sys_frames import Frame
from apipeline.frames.data_frames import TextFrame
from apipeline.pipeline.pipeline import Pipeline, FrameProcessor, FrameDirection

from src.common import interface
from src.common.factory import EngineClass, EngineFactory
from src.core.llm import LLMEnvInit
from src.types.speech.language import TRANSLATE_LANGUAGE
from src.types.frames import TranslationStreamingFrame, TranslationFrame
from src.processors.session_processor import SessionProcessor
from src.common.session import Session, SessionCtx

# use string template to replace placeholders, maybe use jinja2 like tokenizer chat_template with tpl logic
PROMPT_TPL_MAP = {
    "seed-x": "Translate the following $SRC_LANGUAGE sentence into $TARGET_LANGUAGE:\n$TEXT <$TARGET>",
    "hunyuan-mt": "<|startoftext|>Translate the following segment into $TARGET_LANGUAGE, without additional explanation.\n\n$TEXT<|extra_0|>",
    "hunyuan-mt-zh": "<|startoftext|>把下面的文本翻译成$TARGET_LANGUAGE，不要额外解释。\n\n$TEXT<|extra_0|>",
}


class LLMTranslateProcessor(SessionProcessor):
    """
    LLM translate processor with transformers tokenizer
    """

    def __init__(
        self,
        tokenizer: AutoTokenizer,
        generator: interface.ILlmGenerator | EngineClass,
        session: Session | None = None,
        src: str = "en",
        target: str = "zh",
        streaming: bool = False,
        prompt_tpl: str = "seed-x",
        **kwargs,
    ):
        super().__init__(session, **kwargs)
        assert tokenizer is not None, "tokenizer must be provided"
        self.tokenizer = tokenizer
        assert generator is not None, "generator must be provided"
        self.generator = generator
        prompt_tpl = prompt_tpl or "seed-x"
        assert prompt_tpl in PROMPT_TPL_MAP, f"prompt_tpl {prompt_tpl} not supported"

        if src not in TRANSLATE_LANGUAGE:
            raise Exception(f"src language {src} not supported")
        if target not in TRANSLATE_LANGUAGE:
            raise Exception(f"target language {target} not supported")

        self._src = src
        self._target = target
        self._streaming = streaming

        self._translate_frame = TextFrame
        self._translate_tpl = Template(PROMPT_TPL_MAP[prompt_tpl])
        logging.info(f"use {prompt_tpl} translate prompt with streaming: {self._streaming}")

    def set_translate_frame(self, frame_type=TextFrame):
        if not issubclass(frame_type, TextFrame):
            raise Exception(f"frame_type must be a subclass of {TextFrame.__name__}")
        self._translate_frame = frame_type

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)
        if self._src == self._target:
            await self.push_frame(frame, direction)
            return

        if not isinstance(frame, self._translate_frame):
            await self.push_frame(frame, direction)
            return

        await self.handle_translate_frame(frame)

    async def handle_translate_frame(self, frame: TextFrame):
        prompt = self._translate_tpl.substitute(
            SRC=self._src,
            TARGET=self._target,
            SRC_LANGUAGE=TRANSLATE_LANGUAGE[self._src],
            TARGET_LANGUAGE=TRANSLATE_LANGUAGE[self._target],
            TEXT=frame.text,
        )
        logging.info(f"{prompt=}")
        token_ids = self.tokenizer.encode(prompt)
        self.session.ctx.state["token_ids"] = token_ids
        if "ctranslate2" in self.generator.SELECTED_TAG:
            start_tokens = self.tokenizer.convert_ids_to_tokens(token_ids)
            self.session.ctx.state["tokens"] = start_tokens

        await self.start_processing_metrics()
        await self.start_ttfb_metrics()
        first = True
        gen_token_ids = []
        async for gen_token_id in self.generator.generate(self.session, max_new_tokens=512):
            if first:
                await self.stop_ttfb_metrics()
                first = False

            if gen_token_id == self.tokenizer.bos_token_id:
                continue
            if gen_token_id == self.tokenizer.eos_token_id:
                break

            if self._streaming:
                gen_text = self.tokenizer.decode(gen_token_id)
                await self.queue_frame(TranslationStreamingFrame(text=gen_text))
            gen_token_ids.append(gen_token_id)

        if self._streaming:
            await self.queue_frame(TranslationStreamingFrame(text="", is_final=True))

        await self.stop_processing_metrics()

        if self._streaming:
            return

        text = self.tokenizer.decode(gen_token_ids)
        await self.push_frame(
            TranslationFrame(
                src_lang=self._src,
                target_lang=self._target,
                src_text=frame.text,
                text=text,
            )
        )



import torch
from torch import nn
from torch.nn import functional as F

from .conv import Conv2d

logloss = nn.BCELoss(reduction="none")
def cosine_loss(a, v, y):
    d = nn.functional.cosine_similarity(a, v)
    d = d.clamp(0,1) # cosine_similarity的取值范围是【-1，1】，BCE如果输入负数会报错RuntimeError: CUDA error: device-side assert triggered
    loss = logloss(d.unsqueeze(1), y).squeeze()
    loss = loss.mean()
    return loss, d

def get_sync_loss(
    audio_embed, 
    gt_frames, 
    pred_frames, 
    syncnet, 
    adapted_weight,
    frames_left_index=0,
    frames_right_index=16,
):
    # 跟gt_frames做随机的插入交换，节省显存开销
    assert pred_frames.shape[1] == (frames_right_index - frames_left_index) * 3
    # 3通道图像
    frames_sync_loss = torch.cat(
        [gt_frames[:, :3 * frames_left_index, ...], pred_frames, gt_frames[:, 3 * frames_right_index:, ...]], 
        axis=1
    )
    vision_embed = syncnet.get_image_embed(frames_sync_loss)
    y = torch.ones(frames_sync_loss.size(0), 1).float().to(audio_embed.device)
    loss, score = cosine_loss(audio_embed, vision_embed, y)
    return loss, score

class SyncNet_color(nn.Module):
    def __init__(self):
        super(SyncNet_color, self).__init__()

        self.face_encoder = nn.Sequential(
            Conv2d(15, 32, kernel_size=(7, 7), stride=1, padding=3),

            Conv2d(32, 64, kernel_size=5, stride=(1, 2), padding=1),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(64, 128, kernel_size=3, stride=2, padding=1),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(128, 256, kernel_size=3, stride=2, padding=1),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(256, 512, kernel_size=3, stride=2, padding=1),
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(512, 512, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(512, 512, kernel_size=3, stride=2, padding=1),
            Conv2d(512, 512, kernel_size=3, stride=1, padding=0),
            Conv2d(512, 512, kernel_size=1, stride=1, padding=0),)

        self.audio_encoder = nn.Sequential(
            Conv2d(1, 32, kernel_size=3, stride=1, padding=1),
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(32, 32, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(32, 64, kernel_size=3, stride=(3, 1), padding=1),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(64, 64, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(64, 128, kernel_size=3, stride=3, padding=1),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(128, 128, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(128, 256, kernel_size=3, stride=(3, 2), padding=1),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),
            Conv2d(256, 256, kernel_size=3, stride=1, padding=1, residual=True),

            Conv2d(256, 512, kernel_size=3, stride=1, padding=0),
            Conv2d(512, 512, kernel_size=1, stride=1, padding=0),)

    def forward(self, audio_sequences, face_sequences): # audio_sequences := (B, dim, T)
        face_embedding = self.face_encoder(face_sequences)
        audio_embedding = self.audio_encoder(audio_sequences)

        audio_embedding = audio_embedding.view(audio_embedding.size(0), -1)
        face_embedding = face_embedding.view(face_embedding.size(0), -1)

        audio_embedding = F.normalize(audio_embedding, p=2, dim=1)
        face_embedding = F.normalize(face_embedding, p=2, dim=1)


        return audio_embedding, face_embedding